# Unix SMB/CIFS implementation.
#
# Tests for samba-tool commands for Key Distribution Services
#
# Copyright © Catalyst.Net Ltd. 2024
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import re
from datetime import datetime, timezone

from .base import SambaToolCmdTest
from samba.dcerpc import misc

from samba.nt_time import (nt_now,
                           NT_TICKS_PER_SEC,
                           nt_time_from_string,
                           string_from_nt_time)

from ldb import SCOPE_SUBTREE, Dn

from samba.tests.gkdi import create_root_key


HOST = "ldap://{DC_SERVER}".format(**os.environ)
CREDS = "-U{DC_USERNAME}%{DC_PASSWORD}".format(**os.environ)
SMBCONF = os.environ['SERVERCONFFILE']

# alice%Secret007
NON_ADMIN_CREDS = "-U{DOMAIN_USER}%{DOMAIN_USER_PASSWORD}".format(**os.environ)

TIMESTAMP_RE = r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}\+00:00'

NOWISH = 'about now'


class KdsRootKeyTestsBase(SambaToolCmdTest):
    @classmethod
    def setUpClass(cls):
        cls.samdb = cls.getSamDB("-H", HOST, CREDS)
        dn = cls.samdb.get_config_basedn()
        dn.add_child("CN=Master Root Keys,CN=Group Key Distribution Service,CN=Services")
        cls.root_key_base_dn = dn

        # we'll add one for all tests to rely on -- but most will add
        # their own.
        super().setUpClass()

    @classmethod
    def _create_root_key_timediff(cls, create_diff=0, use_diff=0):
        now = nt_now()
        nt_create = now + create_diff * NT_TICKS_PER_SEC
        nt_use = now + use_diff * NT_TICKS_PER_SEC
        guid, dn = create_root_key(cls.samdb,
                                   cls.root_key_base_dn,
                                   current_nt_time=nt_create,
                                   use_start_time=nt_use)

        return guid, dn, nt_create, nt_use

    def _create_root_key_timediff_cleanup(self, create_diff=0, use_diff=0):
        """create a root key that will disappear when the test ends."""
        guid, dn, nt_create, nt_use = self._create_root_key_timediff(
            create_diff,
            use_diff)
        self.addCleanup(self.samdb.delete, dn)
        return guid, dn, nt_create, nt_use

    def _check_timestamp(self, isotimestamp, expected, range=10000):
        """Check that a timestamp string matches an nt-time.

        By default we give a millisecond of leeway, because the ISO
        timestamp has less resolution than NT time (at most 6 decimal
        digits for seconds).
        """

        t = nt_time_from_string(isotimestamp)

        if expected is None:
            # we don't know what we want, but at least it's a time!
            return

        if expected is NOWISH:
            expected = nt_now()
            range = 2.0 * NT_TICKS_PER_SEC

        self.assertGreaterEqual(t, expected - range)
        self.assertLessEqual(t, expected + range)

    def _test_list_output_snippet(self, output,
                                  guid=r'\b[0-9a-fA-F-]{36}\b',
                                  created=None,
                                  used_from=None,
                                  verbose=False):
        # name 1146a853-b604-75ac-5acc-4ef4f0530584
        #    created        2024-02-15T22:55:47.865576+00:00 (about 4 days ago)
        #    usable from    2024-02-15T22:55:47.865576+00:00 (about 4 days ago)
        self.assertRegex(output, f"(?m)^name {guid}$")

        m = re.search(f' created +({TIMESTAMP_RE})', output)
        self.assertIsNotNone(m, "create timestamp not found")
        create_timestamp = m.group(1)
        self._check_timestamp(create_timestamp, created)

        m = re.search(f' usable from +({TIMESTAMP_RE})', output)
        self.assertIsNotNone(m, "usable from timestamp not found")
        used_from_timestamp = m.group(1)
        self._check_timestamp(used_from_timestamp, used_from)

        if verbose:
            dn = f"CN={guid},{self.root_key_base_dn}"
            self.assertRegex(output, f"(?m)^ +dn +{dn}$")
            self.assertRegex(output, r"(?m)^ +whenCreated +\d{14}.0Z$")
            self.assertRegex(output, r"(?m)^ +whenChanged +\d{14}.0Z$")
            self.assertRegex(output, r"(?m)^ +objectGUID +[0-9a-fA-F-]{36}$")
            self.assertRegex(output, r"(?m)^ +msKds-KDFAlgorithmID \w+$")
            self.assertRegex(output, r"(?m)^ +msKds-KDFParam \w+$")
            self.assertRegex(output, r"(?m)^ +msKds-SecretAgreementAlgorithmID \w+$")
            self.assertRegex(output, r"(?m)^ +msKds-PublicKeyLength \d+$")
            self.assertRegex(output, r"(?m)^ +msKds-PrivateKeyLength \d+$")
            self.assertRegex(output, r"(?m)^ +msKds-Version  1$")
            self.assertRegex(output, rf"(?m)^ +msKds-DomainID [\w=, ]+{self.samdb.domain_dn()}$",
                             re.MULTILINE)
            self.assertRegex(output, f"(?m)^ +cn +{guid}$")  # same guid as name

    def _test_list_output_json_snippet(self, snippet,
                                       guid=r'\b[0-9a-fA-F-]{36}\b',
                                       created=None,
                                       used_from=None,
                                       verbose=False):

        _guid = lambda x: re.fullmatch(str(guid), x)
        _hexstr = lambda x: re.fullmatch('[0-9a-fA-F]+', x)
        _str = lambda x: isinstance(x, str)
        _int = lambda x: isinstance(x, int)

        # these next 2 will raise an assertion error on failure
        def _used_from(x):
            self._check_timestamp(x, used_from)
            return True

        def _created(x):
            self._check_timestamp(x, used_from)
            return True

        validators = {
            "cn": _guid,
            "dn": _str,
            "msKds-CreateTime": _created,
            "msKds-DomainID": _str,
            "msKds-KDFAlgorithmID": _str,
            "msKds-KDFParam": _hexstr,
            "msKds-PrivateKeyLength": _int,
            "msKds-PublicKeyLength": _int,
            "msKds-SecretAgreementAlgorithmID": _str,
            "msKds-UseStartTime": _used_from,
            "msKds-Version": _int,
            "name": _guid,
            "objectGUID": _str,
            "whenChanged": _str,
            "whenCreated": _str,
        }
        if verbose:
            keys = validators
        else:
            keys = ["name", "msKds-UseStartTime", "msKds-CreateTime", "dn"]

        self.assertEqual(len(keys), len(snippet), f"keys: {keys}, json: {snippet}")

        for k in keys:
            f = validators.get(k)
            v = snippet.get(k)
            self.assertTrue(f(v), f"{k} value {v} is wrong or malformed")

    def _get_root_key_guids(self):
        """Get the current list of GUIDs."""
        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        return [x['name'] for x in json.loads(out)]

    def _delete_root_key(self, guid):
        dn = Dn(self.samdb, str(self.root_key_base_dn))
        dn.add_child(f"CN={guid}")
        self.samdb.delete(dn)

class KdsRootKeyTests(KdsRootKeyTestsBase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # we'll add one for all tests to rely on.
        cls.common_guid, cls.common_dn, cls.common_time, _ = cls._create_root_key_timediff()
        cls.addClassCleanup(cls.samdb.delete, cls.common_dn)

    def test_list(self):

        """Do we list root keys with the expected info?"""
        # For this test we also need to create some root keys.
        guid, dn, _created, _used = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "list",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        # the output looks something like
        #
        #------------------------------------------------------------------------
        # 2 root keys found.
        #
        # name d58e85d7-ffc4-d118-9c43-46fac38dea05
        #   created        2024-02-27T09:09:21.065486+00:00 (about 1 seconds ago)
        #   usable from    2024-02-27T09:09:21.065486+00:00 (about 1 seconds ago)
        #
        # name 8f3e6557-3ec9-cb84-2ecd-9e258df68e79
        #   created        2024-02-27T09:09:10.853494+00:00 (about 12 seconds ago)
        #   usable from    2024-02-27T09:09:10.853494+00:00 (about 12 seconds ago)
        #-------------------------------------------------------------------------
        #
        # we want to check the various bits.

        parts = out.rstrip().split("\n\n")

        self.assertEqual(parts[0], f"{len(parts) - 1} root keys found.")

        self._test_list_output_snippet(parts[1], guid,
                                       created=NOWISH,
                                       used_from=NOWISH)

        guid2, dn2, _created2, _used2 = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "list",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        parts2 = out.rstrip().split("\n\n")
        self.assertEqual(parts2[0], f"{len(parts)} root keys found.")
        self.assertEqual(len(parts2), len(parts) + 1)

        # we want to check that both of them are still there, in the
        # right order, which is newest first.
        self._test_list_output_snippet(parts2[1], guid2,
                                       created=_created2,
                                       used_from=_used2)
        self._test_list_output_snippet(parts2[2], guid,
                                       created=_created,
                                       used_from=_used)

    def test_list_verbose(self):
        """Do we list root keys with the expected info?"""
        guid, dn, _created, _used = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "-v",
                                       "-H", HOST, CREDS)

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        self._test_list_output_snippet(out, guid, verbose=True)

        guid2, dn2, _created2, _used2 = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "-v",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        self._test_list_output_snippet(out, guid2, verbose=True)

        # in case there are other root keys, we will test each piece
        # using the default '[0-9a-fA-F-]{36}' guid-ish assertion.

        pieces = out.rstrip().split('\n\n')
        self.assertRegex(pieces[0], f'{len(pieces) - 1} root keys found.')

        for piece in pieces[1:]:
            self._test_list_output_snippet(piece, verbose=True)

    def test_list_json(self):
        """The JSON should be a list of dicts, containing the right things"""
        guid, dn, _created, _used = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "-v", "--json",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        for snippet in data:
            self._test_list_output_json_snippet(snippet, verbose=True)

        # non-verbose
        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        for snippet in data:
            self._test_list_output_json_snippet(snippet)

    def test_view_key_that_exists(self):
        guid, dn, _created, _used = self._create_root_key_timediff_cleanup()
        cmd = ["domain", "kds", "root-key", "view",
               "-H", HOST, CREDS,
               "--name", str(guid)]

        result, out, err = self.runcmd(*cmd)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        self._test_list_output_snippet(out, guid,
                                       created=NOWISH,
                                       used_from=NOWISH,
                                       verbose=True)

    def test_view_key_that_exists_json(self):
        guid, dn, _created, _used = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "--json",
                                       "--name", str(guid),
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        self._test_list_output_json_snippet(data, guid,
                                            created=_created,
                                            used_from=_used,
                                            verbose=True)


    def test_view_key_latest_json(self):
        guid, dn, _created, _used = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "--json",
                                       "--latest",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        self._test_list_output_json_snippet(data, guid,
                                            created=_created,
                                            used_from=_used,
                                            verbose=True)

        # if we make a new now-ish key, it will be shown with
        # --latest, forgetting the old one.
        guid2, dn2, _created2, _used2 = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "--json",
                                       "--latest",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        self._test_list_output_json_snippet(data, guid2,
                                            created=_created2,
                                            used_from=_used2,
                                            verbose=True)

        # if we make a new backdated key, it will not be shown as
        # latest, even though it was the most recently created.

        self._create_root_key_timediff_cleanup(use_diff=-600)

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "--json",
                                       "--latest",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        self._test_list_output_json_snippet(data, guid2,
                                            created=_created2,
                                            used_from=_used2,
                                            verbose=True)

        # if we make a future-dated key, it will be shown as
        # latest, even though it doesn't work yet.

        guid3, dn3, _created3, _used3 = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "--json",
                                       "--latest",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        self._test_list_output_json_snippet(data, guid3,
                                            created=_created3,
                                            used_from=_used3,
                                            verbose=True)

    def test_view_non_existent(self):
        """Viewing a non-existent GUID should fail, regardless of what exists."""
        guid = misc.GUID(b'a' * 16)

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "-H", HOST, CREDS,
                                       "--name", str(guid))
        self.assertCmdFail(result)

        self.assertIn("ERROR: no such root key: 61616161-6161-6161-6161-616161616161",
                      err)

    def test_view_non_existent_json(self):
        guid = misc.GUID(b'a' * 16)

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "-H", HOST, CREDS,
                                       "--name", str(guid),
                                       "--json")
        self.assertCmdFail(result)
        data = json.loads(out)
        self.assertEqual(
            data,
            {
                "message": f"no such root key: {guid}",
                "status": "error"
            })

    def test_delete_non_existent(self):
        """Deletion of non-existent guid should fail"""
        guid = 'eeeeeeee-1111-eeee-1111-000000000000'
        result, out, err = self.runcmd("domain", "kds", "root-key", "delete",
                                       "-H", HOST, CREDS,
                                       "--name", guid)
        self.assertCmdFail(result)
        self.assertIn(f"ERROR: no such root key: {guid}", err)

    def test_delete_non_existent_json(self):
        """Deletion of non-existent guids should fail"""
        for guid in ('eeeeeeee-1111-eeee-1111-000000000000',
                     'foo',
                     ''):
            result, out, err = self.runcmd("domain", "kds", "root-key", "delete",
                                           "-H", HOST, CREDS,
                                           "--name", guid,
                                           "--json")
            self.assertCmdFail(result)
            data = json.loads(out)
            self.assertEqual(
                data,
                {
                    "message": f"no such root key: {guid}",
                    "status": "error"
                })

    def test_create(self):
        """does create work?"""
        pre_create = self._get_root_key_guids()

        result, out, err = self.runcmd("domain", "kds", "root-key", "create",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        post_create = self._get_root_key_guids()

        new_guids = list(set(post_create) - set(pre_create))
        gone_guids = set(pre_create) - set(post_create)
        self.assertEqual(len(gone_guids), 0)
        self.assertEqual(len(new_guids), 1)
        self.assertRegex(out,
                         f"created root key {new_guids[0]}, usable from {TIMESTAMP_RE}")
        self._delete_root_key(new_guids[0])

    def test_create_json(self):
        """does create work?"""
        pre_create = self._get_root_key_guids()

        result, out, err = self.runcmd("domain", "kds", "root-key", "create",
                                       "-H", HOST, CREDS, "--json")
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        post_create = self._get_root_key_guids()

        new_guids = list(set(post_create) - set(pre_create))
        gone_guids = set(pre_create) - set(post_create)
        self.assertEqual(len(gone_guids), 0)
        self.assertEqual(len(new_guids), 1)
        data = json.loads(out)
        self.assertEqual(data['dn'], f"CN={new_guids[0]},{self.root_key_base_dn}")
        self.assertEqual(data['status'], 'OK')
        self.assertRegex(data['message'],
                         f"created root key {new_guids[0]}, usable from {TIMESTAMP_RE}")
        self._delete_root_key(new_guids[0])

    def test_create_json_non_admin(self):
        """can you create a root-key without being admin?"""
        pre_create = self._get_root_key_guids()

        result, out, err = self.runcmd("domain", "kds", "root-key", "create",
                                       "-H", HOST, NON_ADMIN_CREDS, "--json")
        self.assertCmdFail(result)

        post_create = self._get_root_key_guids()

        self.assertEqual(set(pre_create), set(post_create))
        data = json.loads(out)
        self.assertEqual(data['status'], 'error')
        self.assertEqual(data['message'], 'User has insufficient access rights')
        self.assertEqual(err, "", "not expecting stderr messages")

    def test_create_json_1997(self):
        """does create work?"""
        pre_create = self._get_root_key_guids()

        result, out, err = self.runcmd("domain", "kds", "root-key", "create",
                                       "-H", HOST, CREDS, "--json",
                                       "--use-start-time",
                                       "1997-11-11T23:18:00.259810+00:00")
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        post_create = self._get_root_key_guids()

        new_guids = list(set(post_create) - set(pre_create))
        gone_guids = set(pre_create) - set(post_create)
        self.assertEqual(len(gone_guids), 0)
        self.assertEqual(len(new_guids), 1)
        data = json.loads(out)
        self.assertEqual(data['dn'], f"CN={new_guids[0]},{self.root_key_base_dn}")
        self.assertEqual(data['status'], 'OK')
        self.assertRegex(data['message'],
                         f"created root key {new_guids[0]}, usable from 1997-11-1")
        self._delete_root_key(new_guids[0])

    def test_create_json_2197(self):
        """does create work?"""
        pre_create = self._get_root_key_guids()

        result, out, err = self.runcmd("domain", "kds", "root-key", "create",
                                       "-H", HOST, CREDS, "--json",
                                       "--use-start-time",
                                       "2197-11-11T23:18:00")
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        post_create = self._get_root_key_guids()

        new_guids = list(set(post_create) - set(pre_create))
        gone_guids = set(pre_create) - set(post_create)
        self.assertEqual(len(gone_guids), 0)
        self.assertEqual(len(new_guids), 1)
        data = json.loads(out)
        self.assertEqual(data['dn'], f"CN={new_guids[0]},{self.root_key_base_dn}")
        self.assertEqual(data['status'], 'OK')
        self.assertRegex(data['message'],
                         f"created root key {new_guids[0]}, usable from 2197-11-1")
        self._delete_root_key(new_guids[0])

    def test_create_future(self):
        """does create work, with a use-start-time 500 seconds in the
        future?"""
        pre_create = self._get_root_key_guids()
        now = nt_now()
        later = now + 500 * NT_TICKS_PER_SEC
        timestamp = string_from_nt_time(later)

        result, out, err = self.runcmd("domain", "kds", "root-key", "create",
                                       "-H", HOST, CREDS, "--json",
                                       "--use-start-time", timestamp)

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")

        post_create = self._get_root_key_guids()

        new_guids = list(set(post_create) - set(pre_create))
        gone_guids = set(pre_create) - set(post_create)
        self.assertEqual(len(gone_guids), 0)
        self.assertEqual(len(new_guids), 1)
        data = json.loads(out)
        self.assertEqual(data['dn'], f"CN={new_guids[0]},{self.root_key_base_dn}")
        self.assertEqual(data['status'], 'OK')
        self.assertRegex(data['message'],
                         f"created root key {new_guids[0]}, usable from {timestamp[:-10]}")
        self._delete_root_key(new_guids[0])

    def test_delete(self):
        """does delete work?"""
        # make one to delete, and get the list as JSON
        _guid, dn, _created, _used = self._create_root_key_timediff()
        guid = str(_guid)

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        pre_delete = json.loads(out)

        result, out, err = self.runcmd("domain", "kds", "root-key", "delete",
                                       "-H", HOST, CREDS,
                                       "--name", guid)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        self.assertEqual(out, f"deleted root key {guid}\n")

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        post_delete = json.loads(out)

        self.assertEqual(len(pre_delete), len(post_delete) + 1)

        post_names = [x['name'] for x in post_delete]
        pre_names = [x['name'] for x in pre_delete]

        self.assertIn(guid, pre_names)
        self.assertNotIn(guid, post_names)

    def test_delete_json(self):
        """does delete --json work?"""
        _guid, dn, _created, _used = self._create_root_key_timediff()
        guid = str(_guid)

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        pre_delete = json.loads(out)

        result, out, err = self.runcmd("domain", "kds", "root-key", "delete",
                                       "-H", HOST, CREDS, "--json",
                                       "--name", guid)

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        self.assertEqual(
            data,
            {
                "message": f"deleted root key {guid}",
                "status": "error"
            })

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        post_delete = json.loads(out)

        self.assertEqual(len(pre_delete), len(post_delete) + 1)

        post_names = [x['name'] for x in post_delete]
        pre_names = [x['name'] for x in pre_delete]

        self.assertIn(guid, pre_names)
        self.assertNotIn(guid, post_names)

    def test_delete_non_admin(self):
        """does delete as non-admin fail?"""
        # make one to delete, and get the list as JSON
        _guid, dn, _created, _used = self._create_root_key_timediff()
        guid = str(_guid)

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        pre_delete = json.loads(out)

        result, out, err = self.runcmd("domain", "kds", "root-key", "delete",
                                       "-H", HOST, NON_ADMIN_CREDS,
                                       "--name", guid)
        self.assertCmdFail(result)
        self.assertIn(f"ERROR: no such root key: {guid}", err)

        # a bad guid should be just like a good guid
        guid2 = 'eeeeeeee-1111-eeee-1111-000000000000'
        result, out2, err2 = self.runcmd("domain", "kds", "root-key", "delete",
                                         "-H", HOST, NON_ADMIN_CREDS,
                                         "--name", guid2)
        self.assertCmdFail(result)
        self.assertIn(f"ERROR: no such root key: {guid2}", err2)

        result, out, err = self.runcmd("domain", "kds", "root-key", "list", "--json",
                                       "-H", HOST, CREDS)
        post_delete = json.loads(out)

        self.assertEqual(len(pre_delete), len(post_delete))

        post_names = [x['name'] for x in post_delete]
        pre_names = [x['name'] for x in pre_delete]

        self.assertIn(guid, pre_names)
        self.assertIn(guid, post_names)

    def test_list_non_admin(self):
        """There are root keys, but non-admins can't see them"""
        result, out, err = self.runcmd("domain", "kds", "root-key", "list",
                                       "-H", HOST, NON_ADMIN_CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        self.assertEqual(out, "no root keys found.\n")

    def test_list_json_non_admin(self):
        """Insufficient rights should look like an empty list."""
        # this is a copy of the KdsNoRootKeyTests test below --
        # non-admin should look exactly like an empty list.
        for extra in ([], ["-v"]):
            result, out, err = self.runcmd("domain", "kds", "root-key", "list",
                                           "-H", HOST, NON_ADMIN_CREDS, "--json", *extra)
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "not expecting error messages")
            data = json.loads(out)
            self.assertEqual(data, [])

    def test_view_key_non_admin(self):
        """should not appear to non-admin"""
        guid, dn, _created, _used = self._create_root_key_timediff_cleanup()

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "--json",
                                       "--name", str(guid),
                                       "-H", HOST, NON_ADMIN_CREDS)
        self.assertCmdFail(result)
        self.assertEqual(err, "", "not expecting error messages")
        data = json.loads(out)
        data = json.loads(out)
        self.assertEqual(
            data,
            {
                "message": f"no such root key: {guid}",
                "status": "error"
            })


class KdsNoRootKeyTests(KdsRootKeyTestsBase):
    """Here we test the case were there are no root keys, which we need to
    ensure by deleting any that are there.
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # We delete all the root keys, and add one back at the end,
        # in case other tests want there to be one.
        res = cls.samdb.search(cls.root_key_base_dn,
                               scope=SCOPE_SUBTREE,
                               expression="(objectClass = msKds-ProvRootKey)")

        for msg in res:
            cls.samdb.delete(msg.dn)

        cls.addClassCleanup(cls.samdb.new_gkdi_root_key)

    def test_list_empty(self):
        """Check the message when there are no root keys"""
        result, out, err = self.runcmd("domain", "kds", "root-key", "list",
                                       "-H", HOST, CREDS)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "not expecting error messages")
        self.assertEqual(out, "no root keys found.\n")

    def test_list_empty_json(self):
        """The JSON should be an empty list when there are no root keys"""
        # verbose flag makes no difference here.
        for extra in ([], ["-v"]):
            result, out, err = self.runcmd("domain", "kds", "root-key", "list",
                                           "-H", HOST, CREDS, "--json", *extra)
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "not expecting error messages")
            data = json.loads(out)
            self.assertEqual(data, [])

    def test_list_empty_json_non_admin(self):
        """Insufficient rights should look like an empty list."""
        # verbose flag makes no difference here.
        for extra in ([], ["-v"]):
            result, out, err = self.runcmd("domain", "kds", "root-key", "list",
                                           "-H", HOST, NON_ADMIN_CREDS, "--json", *extra)
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "not expecting error messages")
            data = json.loads(out)
            self.assertEqual(data, [])

    def test_view_latest_non_existent(self):
        """With no root keys, --latest should return an error"""

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "-H", HOST, CREDS,
                                       "--latest")

        self.assertEqual(err, "ERROR: no root keys found\n")
        self.assertCmdFail(result)

    def test_view_latest_non_existent_json(self):
        """With no root keys, --latest should return an error"""

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "-H", HOST, CREDS,
                                       "--json", "--latest")
        self.assertCmdFail(result)
        data = json.loads(out)
        self.assertEqual(
            data,
            {
                "message": "no root keys found",
                "status": "error"
            })

    def test_view_non_existent(self):
        """Viewing a non-existent GUID should fail, regardless of what exists."""
        guid = misc.GUID(b'b' * 16)

        result, out, err = self.runcmd("domain", "kds", "root-key", "view",
                                       "-H", HOST, CREDS,
                                       "--name", str(guid))
        self.assertCmdFail(result)

        self.assertIn("ERROR: no such root key: 62626262-6262-6262-6262-626262626262",
                      err)



# Integration tests for pycredentials
#
# Copyright (C) Catalyst IT Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from samba.tests import TestCase, delete_force
import os

import samba
from samba.auth import system_session
from samba.credentials import (
    Credentials,
)
from samba.dsdb import (
    UF_WORKSTATION_TRUST_ACCOUNT,
    UF_PASSWD_NOTREQD,
)
from samba.samdb import SamDB

"""KRB5 Integration tests for pycredentials.

Separated from py_credentials so as to allow running against just one
environment so we know the server that we add the user on will be our
KDC

"""

MACHINE_NAME = "krb5credstest"


class PyKrb5CredentialsTests(TestCase):

    def setUp(self):
        super().setUp()

        self.server      = os.environ["SERVER"]
        self.domain      = os.environ["DOMAIN"]
        self.host        = os.environ["SERVER_IP"]
        self.lp          = self.get_loadparm()

        self.credentials = self.get_credentials()

        self.session     = system_session()
        self.ldb = SamDB(url="ldap://%s" % self.host,
                         session_info=self.session,
                         credentials=self.credentials,
                         lp=self.lp)

        self.create_machine_account()

    def tearDown(self):
        super().tearDown()
        delete_force(self.ldb, self.machine_dn)

    def test_get_named_ccache(self):
        name = "MEMORY:py_creds_machine"
        ccache = self.machine_creds.get_named_ccache(self.lp,
                                                     name)
        self.assertEqual(ccache.get_name(), name)

    def test_get_unnamed_ccache(self):
        ccache = self.machine_creds.get_named_ccache(self.lp)
        self.assertIsNotNone(ccache.get_name())

    def test_set_named_ccache(self):
        ccache = self.machine_creds.get_named_ccache(self.lp)

        creds = Credentials()
        creds.set_named_ccache(ccache.get_name())

        ccache2 = creds.get_named_ccache(self.lp)
        self.assertEqual(ccache.get_name(), ccache2.get_name())

    #
    # Create the machine account
    def create_machine_account(self):
        self.machine_pass = samba.generate_random_password(32, 32)
        self.machine_name = MACHINE_NAME
        self.machine_dn = "cn=%s,%s" % (self.machine_name, self.ldb.domain_dn())

        # remove the account if it exists, this will happen if a previous test
        # run failed
        delete_force(self.ldb, self.machine_dn)
        # get unicode str for both py2 and py3
        pass_unicode = self.machine_pass.encode('utf-8').decode('utf-8')
        utf16pw = u'"{0}"'.format(pass_unicode).encode('utf-16-le')
        self.ldb.add({
            "dn": self.machine_dn,
            "objectclass": "computer",
            "sAMAccountName": "%s$" % self.machine_name,
            "userAccountControl":
                str(UF_WORKSTATION_TRUST_ACCOUNT | UF_PASSWD_NOTREQD),
            "unicodePwd": utf16pw})

        self.machine_creds = Credentials()
        self.machine_creds.guess(self.get_loadparm())
        self.machine_creds.set_password(self.machine_pass)
        self.machine_creds.set_username(self.machine_name + "$")
        self.machine_creds.set_workstation(self.machine_name)



# Unix SMB/CIFS implementation.
#
# authentication silos
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from samba.netcmd import SuperCommand

from .policy import cmd_domain_auth_policy
from .silo import cmd_domain_auth_silo


class cmd_domain_auth(SuperCommand):
    """Manage authentication silos and policies on the domain."""

    subcommands = {
        "policy": cmd_domain_auth_policy(),
        "silo": cmd_domain_auth_silo(),
    }



#
# NT Time utility functions.
#
# Copyright (C) Catalyst.Net Ltd 2023
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import datetime
from typing import NewType, Optional
import re


NtTime = NewType("NtTime", int)
NtTimeDelta = NewType("NtTimeDelta", int)

NT_TIME_MAX = NtTime((1 << 64) - 1)

NT_EPOCH = datetime.datetime(1601, 1, 1, 0, 0, 0, 0, tzinfo=datetime.timezone.utc)
NT_TICKS_PER_μSEC = 10
NT_TICKS_PER_SEC = NT_TICKS_PER_μSEC * 1_000_000


def _validate_nt_time(nt_time: NtTime) -> None:
    if not isinstance(nt_time, int):
        raise ValueError(f"{nt_time} is not an integer")
    if not 0 <= nt_time <= NT_TIME_MAX:
        raise ValueError(f"{nt_time} is out of range")


def nt_time_from_datetime(tm: datetime.datetime) -> NtTime:
    time_since_epoch = tm - NT_EPOCH
    nt_time = NtTime(round(time_since_epoch.total_seconds() * NT_TICKS_PER_SEC))
    _validate_nt_time(nt_time)
    return nt_time


def nt_now() -> NtTime:
    dt = datetime.datetime.now(datetime.timezone.utc)
    return nt_time_from_datetime(dt)


def datetime_from_nt_time(nt_time: NtTime) -> datetime.datetime:
    _validate_nt_time(nt_time)
    time_since_epoch = datetime.timedelta(microseconds=nt_time / NT_TICKS_PER_μSEC)
    return NT_EPOCH + time_since_epoch


def nt_time_delta_from_timedelta(dt: datetime.timedelta) -> NtTimeDelta:
    return NtTimeDelta(round(dt.total_seconds() * NT_TICKS_PER_SEC))


def timedelta_from_nt_time_delta(nt_time_delta: NtTimeDelta) -> datetime.timedelta:
    return datetime.timedelta(microseconds=nt_time_delta / NT_TICKS_PER_μSEC)


def nt_time_from_string(s: str) -> NtTime:
    """Convert a subset of ISO 8601 date/time strings, ldap timestamps,
    and the string 'now' into NT time.

    The ldap format is

       YYYYmmddHHMMSS.0Z

    which is 14 digits followed by the fixed string '.0Z'. This is
    used in LDIF and internally by ldb.

    The ISO format is

    YYYY-mm-dd[*HH[:MM[:SS[.fff[fff]]]][+HH:MM[:SS[.ffffff]]]]

    where the '*' can be any character, and the optional last
    '[+HH:MM[:SS[.ffffff]]]' is a timezone offset (use '+00:00' for
    UTC).
    """
    try:
        if s == "now":
            dt = datetime.datetime.now(datetime.timezone.utc)
        elif re.match(r"^\d{14}\.0Z$", s):
            # "20230127223641.0Z"
            dt = datetime.datetime.strptime(s, "%Y%m%d%H%M%S.0Z")
        else:
            dt = datetime.datetime.fromisoformat(s)
    except ValueError:
        raise ValueError(
            "Expected a date in either "
            "ISO8601 'YYYY-MM-DD HH:MM:SS' format, "
            "LDAP timestamp 'YYYYmmddHHMMSS.0Z', "
            "or the literal string 'now'. "
            f" Got '{s}'."
        )

    if dt.tzinfo is None:
        # This is a cursed timestamp with no timezone info. We have to
        # guess or nt_time_from_datetime() will fail. The best guess
        # is the system timezone, which we can get this way:
        dt = dt.astimezone()

    return nt_time_from_datetime(dt)


def string_from_nt_time(nttime: NtTime, format: Optional[str] = None) -> str:
    """Format an NtTime date as a string.

    If format is not provided, an ISO 8601 string is used.
    """
    dt = datetime_from_nt_time(nttime)

    if format is not None:
        return dt.strftime(format)

    return dt.isoformat()



# Unix SMB/CIFS implementation.
#
# Site model.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .fields import BooleanField, DnField, IntegerField
from .model import Model


class Site(Model):
    show_in_advanced_view_only = BooleanField("showInAdvancedViewOnly")
    system_flags = IntegerField("systemFlags", readonly=True)

    # Backlinks
    site_object_bl = DnField("siteObjectBL", readonly=True)

    @staticmethod
    def get_base_dn(samdb):
        """Return the base DN for the Site model.

        :param samdb: SamDB connection
        :return: Dn to use for new objects
        """
        base_dn = samdb.get_config_basedn()
        base_dn.add_child("CN=Sites")
        return base_dn

    @staticmethod
    def get_object_class():
        return "site"



# domain management - domain samba3upgrade
#
# Copyright Matthias Dieter Wallnoefer 2009
# Copyright Andrew Kroeger 2009
# Copyright Jelmer Vernooij 2007-2012
# Copyright Giampaolo Lauria 2011
# Copyright Matthieu Patou <mat@matws.net> 2011
# Copyright Andrew Bartlett 2008-2015
# Copyright Stefan Metzmacher 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .classicupgrade import cmd_domain_classicupgrade


class cmd_domain_samba3upgrade(cmd_domain_classicupgrade):
    __doc__ = cmd_domain_classicupgrade.__doc__

    # This command is present for backwards compatibility only,
    # and should not be shown.

    hidden = True



#!/usr/local/bin/python3.14
# Unix SMB/CIFS implementation.
#
# Tests for Key Trust authentication
#
# Based on pkinit_tests.py
#
# Copyright (C) Gary Lockyer <gary@catalyst.net.nz> 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

from datetime import datetime, timedelta

from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.x509.oid import NameOID

from samba.dcerpc import security
from samba.domain.models import User
from samba.key_credential_link import create_key_credential_link
from samba.tests.krb5 import kcrypto
from samba.tests.krb5.kdc_base_test import KDCBaseTest
from samba.tests.krb5.raw_testcase import PkInit, RawKerberosTest
from samba.tests.krb5.rfc4120_constants import (
    DES_EDE3_CBC,
    KDC_ERR_CLIENT_NOT_TRUSTED,
    NT_PRINCIPAL,
    PADATA_PK_AS_REQ,
)
import samba.tests.krb5.rfc4120_pyasn1 as krb5_asn1

SidType = RawKerberosTest.SidType

global_asn1_print = False
global_hexdump = False


class KeyTrustTests(KDCBaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def setUp(self):
        super().setUp()
        self.do_asn1_print = global_asn1_print
        self.do_hexdump = global_hexdump

    def test_key_trust_user(self):
        """
        Test key trust logon for a normal account
        """
        client_creds = self._get_creds()
        target_creds = self.get_service_creds()
        certificate = self._create_certificate(client_creds)
        public_key = certificate.public_bytes(Encoding.DER)
        self._add_key_cred_link(client_creds, public_key)

        self._pkinit_req(client_creds, target_creds, certificate)

    def test_key_trust_user_mismatched_keys(self):
        """
        Test key trust logon for a normal account, where the certificate
        public key does not match the msDS-KeyCredentialLink value
        """
        client_creds = self._get_creds()
        target_creds = self.get_service_creds()
        certificate = self._create_certificate(client_creds)

        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        public_key = private_key.public_key().public_bytes(
            encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
        )
        self._add_key_cred_link(client_creds, public_key)

        self._pkinit_req(
            client_creds,
            target_creds,
            certificate,
            expect_error=KDC_ERR_CLIENT_NOT_TRUSTED,
        )

    def test_key_trust_user_no_keys(self):
        """
        Test key trust logon for a normal account,
        with no msDS-KeyCredentialLink
        """
        client_creds = self._get_creds()
        target_creds = self.get_service_creds()
        certificate = self._create_certificate(client_creds)

        self._pkinit_req(
            client_creds,
            target_creds,
            certificate,
            expect_error=KDC_ERR_CLIENT_NOT_TRUSTED,
        )

    def test_key_trust_computer(self):
        """
        Test key trust logon with a computer account.
        """
        client_creds = self._get_creds(self.AccountType.COMPUTER)
        target_creds = self.get_service_creds()
        certificate = self._create_certificate(client_creds)
        public_key = certificate.public_bytes(Encoding.DER)
        self._add_key_cred_link(client_creds, public_key)

        self._pkinit_req(client_creds, target_creds, certificate)

    def test_key_trust_computer_mismatched_keys(self):
        """
        Test key trust logon for a computer account, where the certificate
        public key does not match the msDS-KeyCredentialLink value
        """
        client_creds = self._get_creds(self.AccountType.COMPUTER)
        target_creds = self.get_service_creds()
        certificate = self._create_certificate(client_creds)

        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        public_key = private_key.public_key().public_bytes(
            encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
        )
        self._add_key_cred_link(client_creds, public_key)

        self._pkinit_req(
            client_creds,
            target_creds,
            certificate,
            expect_error=KDC_ERR_CLIENT_NOT_TRUSTED,
        )

    def test_key_trust_computer_no_keys(self):
        """
        Test key trust logon with a computer account,
        with no msDS-KeyCredentialLink
        """
        client_creds = self._get_creds(self.AccountType.COMPUTER)
        target_creds = self.get_service_creds()
        certificate = self._create_certificate(client_creds)

        self._pkinit_req(
            client_creds,
            target_creds,
            certificate = certificate,
            expect_error=KDC_ERR_CLIENT_NOT_TRUSTED,
        )

    def _get_creds(self, account_type=KDCBaseTest.AccountType.USER, use_cache=False):
        """
        Return credentials with an account having a UPN for performing
        PK-INIT.

        Modified from the version in python/samba/tests/krb5/pkinit_tests.py
        """
        samdb = self.get_samdb()
        realm = samdb.domain_dns_name().upper()

        opts = {"upn": f"{{account}}.{realm}@{realm}"}

        return self.get_cached_creds(
            account_type=account_type, opts=opts, use_cache=use_cache
        )

    def _add_key_cred_link(self, creds, public_key):
        """
        Update the msDS-KeyCredentialLink for the user specified in creds with
        the supplied public key
        """
        link = create_key_credential_link(self.get_samdb(), creds.get_dn(), public_key)

        user = User.find(self.get_samdb(), creds.get_username())
        self.assertIsNotNone(user)
        if user is not None:
            user.key_credential_link = link
            user.save(self.get_samdb())

    def _create_certificate(self, creds):
        """
        Create a new self signed certificate

        Modified from the version in python/samba/tests/krb5/pkinit_tests.py
        """

        certificate_signature = hashes.SHA256

        user_name = creds.get_username()

        builder = x509.CertificateBuilder()

        # Add the subject name.
        cert_name = f"{user_name}@{creds.get_realm().lower()}"
        subject_name = x509.Name(
            [
                # This name can be anything; it isn’t needed to authorize the
                # user. The SubjectAlternativeName is used for that instead.
                x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
                x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "SambaState"),
                x509.NameAttribute(NameOID.ORGANIZATION_NAME, "SambaSelfTesting"),
                x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, "Users"),
                x509.NameAttribute(NameOID.COMMON_NAME, f"{cert_name}"),
            ]
        )
        builder = builder.subject_name(subject_name)

        # The new certificate is self signed
        builder = builder.issuer_name(subject_name)

        one_day = timedelta(1, 0, 0)

        # Put the certificate start time in the past to avoid issues where the
        # KDC considers the certificate to be invalid due to clock skew. Note
        # that if the certificate predates the existence of the account in AD,
        # Windows will refuse authentication unless a strong mapping is
        # present (in the certificate, or in AD).
        # See https://support.microsoft.com/en-us/topic/kb5014754-certificate-based-authentication-changes-on-windows-domain-controllers-ad2c23b0-15d8-4340-a468-4d4f3b188f16#ID0EFR
        builder = builder.not_valid_before(datetime.today() - one_day)
        builder = builder.not_valid_after(datetime.today() + (one_day * 30))

        builder = builder.serial_number(x509.random_serial_number())

        public_key = creds.get_public_key()
        builder = builder.public_key(public_key)

        # Add the SubjectAlternativeName. Windows uses this to map the account
        # to the certificate.
        id_pkinit_ms_san = x509.ObjectIdentifier(str(krb5_asn1.id_pkinit_ms_san))
        encoded_upn = self.der_encode(creds.get_upn(), asn1Spec=krb5_asn1.MS_UPN_SAN())
        ms_upn_san = x509.OtherName(id_pkinit_ms_san, encoded_upn)
        builder = builder.add_extension(
            x509.SubjectAlternativeName([ms_upn_san]),
            critical=False,
        )

        # The key identifier is used to identify the certificate.
        subject_key_id = x509.SubjectKeyIdentifier.from_public_key(public_key)
        builder = builder.add_extension(
            subject_key_id,
            critical=True,
        )

        # Add the key usages for which this certificate is valid.
        builder = builder.add_extension(
            # Heimdal requires that the certificate be valid for digital
            # signatures.
            x509.KeyUsage(digital_signature=True,
                          content_commitment=False,
                          key_encipherment=False,
                          data_encipherment=False,
                          key_agreement=False,
                          key_cert_sign=False,
                          crl_sign=False,
                          encipher_only=False,
                          decipher_only=False),
                          critical=True,
        )

        certificate = builder.sign(
            private_key=creds.get_private_key(),
            algorithm=certificate_signature(),
            backend=default_backend(),
        )

        return certificate

    def _pkinit_req(
        self,
        creds,
        target_creds,
        certificate,
        expect_error=0,
    ):
        """
        Modified from the version in python/samba/tests/krb5/pkinit_tests.py
        """
        signature_algorithm = krb5_asn1.sha1WithRSAEncryption
        signature_algorithm_id = self.AlgorithmIdentifier_create(signature_algorithm)

        digest_algorithm = krb5_asn1.id_sha1
        digest_algorithm_id = self.AlgorithmIdentifier_create(digest_algorithm)

        private_key = creds.get_private_key()

        preauth_key = private_key

        pk_nonce = self.get_Nonce()

        def generate_pk_padata(_kdc_exchange_dict, _callback_dict, req_body):
            # Suppress unused parameter warnings
            _ = (_kdc_exchange_dict, _callback_dict)

            checksum_blob = self.der_encode(req_body, asn1Spec=krb5_asn1.KDC_REQ_BODY())

            # Calculate the SHA1 checksum over the KDC-REQ-BODY. This checksum
            # is required to be present in the authenticator, and must be SHA1.
            digest = hashes.Hash(hashes.SHA1(), default_backend())
            digest.update(checksum_blob)
            digest = digest.finalize()

            ctime, cusec = self.get_KerberosTimeWithUsec()

            # Create the authenticator, which shows that we had possession of
            # the private key at some point.
            authenticator_obj = self.PKAuthenticator_create(
                cusec, ctime, pk_nonce, pa_checksum=digest
            )

            client_public_value = None

            # An optional set of algorithms supported by the client in
            # decreasing order of preference. For whatever reason, if this
            # field is missing or empty, Windows will respond with a slightly
            # differently encoded ReplyKeyPack, wrapping it first in a
            # ContentInfo structure.
            supported_cms_types = [
                self.AlgorithmIdentifier_create(
                    krb5_asn1.id_pkcs1_sha256WithRSAEncryption
                ),
            ]
            auth_pack_obj = self.AuthPack_create(
                authenticator_obj,
                client_public_value=client_public_value,
                supported_cms_types=supported_cms_types,
            )

            asn1_spec = krb5_asn1.AuthPack
            auth_pack = self.der_encode(auth_pack_obj, asn1Spec=asn1_spec())

            signature_hash = self.hash_from_algorithm(signature_algorithm)
            digest_hash = self.hash_from_algorithm(digest_algorithm)

            digest = hashes.Hash(digest_hash(), default_backend())
            digest.update(auth_pack)
            digest = digest.finalize()

            encap_content_info_obj = self.EncapsulatedContentInfo_create(
                krb5_asn1.id_pkinit_authData, auth_pack
            )

            subject_key_id = certificate.extensions.get_extension_for_oid(
                x509.ExtensionOID.SUBJECT_KEY_IDENTIFIER
            )
            signer_identifier = self.SignerIdentifier_create(
                subject_key_id=subject_key_id.value.digest
            )

            auth_data_attr = self.Attribute_create(
                krb5_asn1.id_pkinit_authData, [auth_pack]
            )

            message_digest = self.der_encode(digest, asn1Spec=krb5_asn1.MessageDigest())

            message_digest_attr = self.Attribute_create(
                krb5_asn1.id_messageDigest, [message_digest]
            )

            signed_attrs = [
                # Note: these attributes are optional.
                auth_data_attr,
                message_digest_attr,
            ]
            encoded_signed_attrs = self.der_encode(
                signed_attrs, asn1Spec=krb5_asn1.CMSAttributes()
            )

            pad = padding.PKCS1v15()
            signed = private_key.sign(
                encoded_signed_attrs, padding=pad, algorithm=signature_hash()
            )

            signer_info = self.SignerInfo_create(
                signer_identifier,
                digest_algorithm_id,
                signature_algorithm_id,
                signed,
                signed_attrs=signed_attrs,
            )

            encoded_cert = certificate.public_bytes(serialization.Encoding.DER)
            decoded_cert = self.der_decode(
                encoded_cert, asn1Spec=krb5_asn1.CertificateChoices()
            )

            signed_auth_pack = self.SignedData_create(
                [digest_algorithm_id],
                encap_content_info_obj,
                signer_infos=[signer_info],
                certificates=[decoded_cert],
                crls=None,
            )

            signed_auth_pack = self.der_encode(
                signed_auth_pack, asn1Spec=krb5_asn1.SignedData()
            )

            pk_as_req = self.PK_AS_REQ_create(signed_auth_pack)

            pa_type = PADATA_PK_AS_REQ
            padata = [self.PA_DATA_create(pa_type, pk_as_req)]

            return padata, req_body

        user_name = creds.get_username()
        cname = self.PrincipalName_create(
            name_type=NT_PRINCIPAL, names=user_name.split("/")
        )

        target_name = target_creds.get_username()
        target_realm = target_creds.get_realm()

        target_sname = self.PrincipalName_create(
            name_type=NT_PRINCIPAL, names=["host", target_name[:-1]]
        )
        expected_sname = self.PrincipalName_create(
            name_type=NT_PRINCIPAL, names=[target_name]
        )

        if expect_error:
            check_error_fn = self.generic_check_kdc_error
            check_rep_fn = None

            expected_sname = target_sname
        else:
            check_error_fn = None
            check_rep_fn = self.generic_check_kdc_rep

        kdc_options = "forwardable,renewable,canonicalize,renewable-ok"
        kdc_options = krb5_asn1.KDCOptions(kdc_options)

        ticket_decryption_key = self.TicketDecryptionKey_from_creds(target_creds)

        expected_groups = None
        unexpected_groups = {
            (
                security.SID_FRESH_PUBLIC_KEY_IDENTITY,
                SidType.EXTRA_SID,
                security.SE_GROUP_DEFAULT_FLAGS,
            )
        }

        kdc_exchange_dict = self.as_exchange_dict(
            creds=creds,
            client_cert=certificate,
            expected_crealm=creds.get_realm(),
            expected_cname=cname,
            expected_srealm=target_realm,
            expected_sname=expected_sname,
            expected_supported_etypes=target_creds.tgs_supported_enctypes,
            expected_groups=expected_groups,
            unexpected_groups=unexpected_groups,
            ticket_decryption_key=ticket_decryption_key,
            generate_padata_fn=generate_pk_padata,
            check_error_fn=check_error_fn,
            check_rep_fn=check_rep_fn,
            check_kdc_private_fn=self.generic_check_kdc_private,
            expected_error_mode=expect_error,
            expected_salt=creds.get_salt(),
            preauth_key=preauth_key,
            kdc_options=str(kdc_options),
            using_pkinit=PkInit.PUBLIC_KEY,
            pk_nonce=pk_nonce,
            expect_edata=False,
            expect_matching_nt_hash_in_pac=True,
        )

        till = self.get_KerberosTime(offset=36000)

        etypes = (
            kcrypto.Enctype.AES256,
            kcrypto.Enctype.RC4,
        )
        # DES-EDE3-CBC is required for public-key PK-INIT to work on Windows.
        etypes += (DES_EDE3_CBC,)

        rep = self._generic_kdc_exchange(
            kdc_exchange_dict,
            cname=cname,
            realm=target_realm,
            sname=target_sname,
            till_time=till,
            etypes=etypes,
        )
        if expect_error:
            self.check_error_rep(rep, expect_error)
            return None

        self.check_as_reply(rep)
        return kdc_exchange_dict


if __name__ == "__main__":
    global_asn1_print = False
    global_hexdump = False
    import unittest

    unittest.main()



# gp_chromium_ext samba gpo policy
# Copyright (C) David Mulder <dmulder@suse.com> 2021
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
from samba.gp.gpclass import gp_pol_ext, gp_file_applier
from samba.dcerpc import misc
from samba.common import get_string
from samba.gp.util.logging import log
from tempfile import NamedTemporaryFile

def parse_entry_data(name, e):
    dict_entries = ['VirtualKeyboardFeatures',
                    'DeviceArcDataSnapshotHours',
                    'RequiredClientCertificateForDevice',
                    'RequiredClientCertificateForUser',
                    'RegisteredProtocolHandlers',
                    'WebUsbAllowDevicesForUrls',
                    'DeviceAutoUpdateTimeRestrictions',
                    'DeviceUpdateStagingSchedule',
                    'DeviceMinimumVersion',
                    'DeviceDisplayResolution',
                    'ExtensionSettings',
                    'KerberosAccounts',
                    'NetworkFileSharesPreconfiguredShares',
                    'NetworkThrottlingEnabled',
                    'TPMFirmwareUpdateSettings',
                    'DeviceOffHours',
                    'ParentAccessCodeConfig',
                    'PerAppTimeLimits',
                    'PerAppTimeLimitsWhitelist',
                    'PerAppTimeLimitsAllowlist',
                    'UsageTimeLimit',
                    'PluginVmImage',
                    'DeviceLoginScreenPowerManagement',
                    'PowerManagementIdleSettings',
                    'ScreenLockDelays',
                    'ScreenBrightnessPercent',
                    'DevicePowerPeakShiftDayConfig',
                    'DeviceAdvancedBatteryChargeModeDayConfig',
                    'PrintingPaperSizeDefault',
                    'AutoLaunchProtocolsFromOrigins',
                    'BrowsingDataLifetime',
                    'DataLeakPreventionRulesList',
                    'DeviceLoginScreenWebUsbAllowDevicesForUrls',
                    'DeviceScheduledUpdateCheck',
                    'KeyPermissions',
                    'ManagedBookmarks',
                    'ManagedConfigurationPerOrigin',
                    'ProxySettings',
                    'SystemProxySettings',
                    'WebAppInstallForceList']
    bools = ['ShowAccessibilityOptionsInSystemTrayMenu',
             'LargeCursorEnabled',
             'SpokenFeedbackEnabled',
             'HighContrastEnabled',
             'VirtualKeyboardEnabled',
             'StickyKeysEnabled',
             'KeyboardDefaultToFunctionKeys',
             'DictationEnabled',
             'SelectToSpeakEnabled',
             'KeyboardFocusHighlightEnabled',
             'CursorHighlightEnabled',
             'CaretHighlightEnabled',
             'MonoAudioEnabled',
             'AccessibilityShortcutsEnabled',
             'AutoclickEnabled',
             'DeviceLoginScreenDefaultLargeCursorEnabled',
             'DeviceLoginScreenDefaultSpokenFeedbackEnabled',
             'DeviceLoginScreenDefaultHighContrastEnabled',
             'DeviceLoginScreenDefaultVirtualKeyboardEnabled',
             'DeviceLoginScreenLargeCursorEnabled',
             'DeviceLoginScreenSpokenFeedbackEnabled',
             'DeviceLoginScreenHighContrastEnabled',
             'DeviceLoginScreenVirtualKeyboardEnabled',
             'DeviceLoginScreenDictationEnabled',
             'DeviceLoginScreenSelectToSpeakEnabled',
             'DeviceLoginScreenCursorHighlightEnabled',
             'DeviceLoginScreenCaretHighlightEnabled',
             'DeviceLoginScreenMonoAudioEnabled',
             'DeviceLoginScreenAutoclickEnabled',
             'DeviceLoginScreenStickyKeysEnabled',
             'DeviceLoginScreenKeyboardFocusHighlightEnabled',
             'DeviceLoginScreenShowOptionsInSystemTrayMenu',
             'DeviceLoginScreenAccessibilityShortcutsEnabled',
             'FloatingAccessibilityMenuEnabled',
             'ArcEnabled',
             'UnaffiliatedArcAllowed',
             'AppRecommendationZeroStateEnabled',
             'DeviceBorealisAllowed',
             'UserBorealisAllowed',
             'SystemUse24HourClock',
             'DefaultSearchProviderEnabled',
             'ChromeOsReleaseChannelDelegated',
             'DeviceAutoUpdateDisabled',
             'DeviceAutoUpdateP2PEnabled',
             'DeviceUpdateHttpDownloadsEnabled',
             'RebootAfterUpdate',
             'BlockExternalExtensions',
             'VoiceInteractionContextEnabled',
             'VoiceInteractionHotwordEnabled',
             'EnableMediaRouter',
             'ShowCastIconInToolbar',
             'DriveDisabled',
             'DriveDisabledOverCellular',
             'DisableAuthNegotiateCnameLookup',
             'EnableAuthNegotiatePort',
             'BasicAuthOverHttpEnabled',
             'AuthNegotiateDelegateByKdcPolicy',
             'AllowCrossOriginAuthPrompt',
             'NtlmV2Enabled',
             'IntegratedWebAuthenticationAllowed',
             'BrowserSwitcherEnabled',
             'BrowserSwitcherKeepLastChromeTab',
             'BrowserSwitcherUseIeSitelist',
             'VirtualMachinesAllowed',
             'CrostiniAllowed',
             'DeviceUnaffiliatedCrostiniAllowed',
             'CrostiniExportImportUIAllowed',
             'CrostiniPortForwardingAllowed',
             'NativeMessagingUserLevelHosts',
             'NetworkFileSharesAllowed',
             'NetBiosShareDiscoveryEnabled',
             'NTLMShareAuthenticationEnabled',
             'DeviceDataRoamingEnabled',
             'DeviceWiFiFastTransitionEnabled',
             'DeviceWiFiAllowed',
             'DeviceAllowBluetooth',
             'DeviceAllowRedeemChromeOsRegistrationOffers',
             'DeviceQuirksDownloadEnabled',
             'SuggestedContentEnabled',
             'DeviceShowLowDiskSpaceNotification',
             'PasswordManagerEnabled',
             'PasswordLeakDetectionEnabled',
             'PluginVmAllowed',
             'PluginVmDataCollectionAllowed',
             'UserPluginVmAllowed',
             'DeviceRebootOnShutdown',
             'PowerManagementUsesAudioActivity',
             'PowerManagementUsesVideoActivity',
             'AllowWakeLocks',
             'AllowScreenWakeLocks',
             'WaitForInitialUserActivity',
             'PowerSmartDimEnabled',
             'DevicePowerPeakShiftEnabled',
             'DeviceBootOnAcEnabled',
             'DeviceAdvancedBatteryChargeModeEnabled',
             'DeviceUsbPowerShareEnabled',
             'PrintingEnabled',
             'CloudPrintProxyEnabled',
             'PrintingSendUsernameAndFilenameEnabled',
             'CloudPrintSubmitEnabled',
             'DisablePrintPreview',
             'PrintHeaderFooter',
             'PrintPreviewUseSystemDefaultPrinter',
             'UserNativePrintersAllowed',
             'UserPrintersAllowed',
             'DeletePrintJobHistoryAllowed',
             'DeviceLoginScreenPrivacyScreenEnabled',
             'PrivacyScreenEnabled',
             'PinUnlockWeakPinsAllowed',
             'PinUnlockAutosubmitEnabled',
             'RemoteAccessHostFirewallTraversal',
             'RemoteAccessHostRequireCurtain',
             'RemoteAccessHostAllowClientPairing',
             'RemoteAccessHostAllowRelayedConnection',
             'RemoteAccessHostAllowUiAccessForRemoteAssistance',
             'RemoteAccessHostAllowFileTransfer',
             'RemoteAccessHostAllowRemoteAccessConnections',
             'AttestationEnabledForUser',
             'SafeBrowsingEnabled',
             'SafeBrowsingExtendedReportingEnabled',
             'DeviceGuestModeEnabled',
             'DeviceAllowNewUsers',
             'DeviceShowUserNamesOnSignin',
             'DeviceEphemeralUsersEnabled',
             'DeviceShowNumericKeyboardForPassword',
             'DeviceFamilyLinkAccountsAllowed',
             'ShowHomeButton',
             'HomepageIsNewTabPage',
             'DeviceMetricsReportingEnabled',
             'DeviceWilcoDtcAllowed',
             'AbusiveExperienceInterventionEnforce',
             'AccessibilityImageLabelsEnabled',
             'AdditionalDnsQueryTypesEnabled',
             'AdvancedProtectionAllowed',
             'AllowDeletingBrowserHistory',
             'AllowDinosaurEasterEgg',
             'AllowFileSelectionDialogs',
             'AllowScreenLock',
             'AllowSyncXHRInPageDismissal',
             'AlternateErrorPagesEnabled',
             'AlwaysOpenPdfExternally',
             'AppCacheForceEnabled',
             'AudioCaptureAllowed',
             'AudioOutputAllowed',
             'AudioProcessHighPriorityEnabled',
             'AudioSandboxEnabled',
             'AutoFillEnabled',
             'AutofillAddressEnabled',
             'AutofillCreditCardEnabled',
             'AutoplayAllowed',
             'BackgroundModeEnabled',
             'BlockThirdPartyCookies',
             'BookmarkBarEnabled',
             'BrowserAddPersonEnabled',
             'BrowserGuestModeEnabled',
             'BrowserGuestModeEnforced',
             'BrowserLabsEnabled',
             'BrowserNetworkTimeQueriesEnabled',
             'BuiltInDnsClientEnabled',
             'CECPQ2Enabled',
             'CaptivePortalAuthenticationIgnoresProxy',
             'ChromeCleanupEnabled',
             'ChromeCleanupReportingEnabled',
             'ChromeOsLockOnIdleSuspend',
             'ClickToCallEnabled',
             'CloudManagementEnrollmentMandatory',
             'CloudPolicyOverridesPlatformPolicy',
             'CloudUserPolicyMerge',
             'CommandLineFlagSecurityWarningsEnabled',
             'ComponentUpdatesEnabled',
             'DNSInterceptionChecksEnabled',
             'DataLeakPreventionReportingEnabled',
             'DefaultBrowserSettingEnabled',
             'DefaultSearchProviderContextMenuAccessAllowed',
             'DeveloperToolsDisabled',
             'DeviceAllowMGSToStoreDisplayProperties',
             'DeviceDebugPacketCaptureAllowed',
             'DeviceLocalAccountManagedSessionEnabled',
             'DeviceLoginScreenPrimaryMouseButtonSwitch',
             'DevicePciPeripheralDataAccessEnabled',
             'DevicePowerwashAllowed',
             'DeviceSystemWideTracingEnabled',
             'Disable3DAPIs',
             'DisableSafeBrowsingProceedAnyway',
             'DisableScreenshots',
             'EasyUnlockAllowed',
             'EditBookmarksEnabled',
             'EmojiSuggestionEnabled',
             'EnableDeprecatedPrivetPrinting',
             'EnableOnlineRevocationChecks',
             'EnableSyncConsent',
             'EnterpriseHardwarePlatformAPIEnabled',
             'ExternalProtocolDialogShowAlwaysOpenCheckbox',
             'ExternalStorageDisabled',
             'ExternalStorageReadOnly',
             'ForceBrowserSignin',
             'ForceEphemeralProfiles',
             'ForceGoogleSafeSearch',
             'ForceMaximizeOnFirstRun',
             'ForceSafeSearch',
             'ForceYouTubeSafetyMode',
             'FullscreenAlertEnabled',
             'FullscreenAllowed',
             'GloballyScopeHTTPAuthCacheEnabled',
             'HardwareAccelerationModeEnabled',
             'HideWebStoreIcon',
             'ImportAutofillFormData',
             'ImportBookmarks',
             'ImportHistory',
             'ImportHomepage',
             'ImportSavedPasswords',
             'ImportSearchEngine',
             'IncognitoEnabled',
             'InsecureFormsWarningsEnabled',
             'InsecurePrivateNetworkRequestsAllowed',
             'InstantTetheringAllowed',
             'IntensiveWakeUpThrottlingEnabled',
             'JavascriptEnabled',
             'LacrosAllowed',
             'LacrosSecondaryProfilesAllowed',
             'LockScreenMediaPlaybackEnabled',
             'LoginDisplayPasswordButtonEnabled',
             'ManagedGuestSessionPrivacyWarningsEnabled',
             'MediaRecommendationsEnabled',
             'MediaRouterCastAllowAllIPs',
             'MetricsReportingEnabled',
             'NTPCardsVisible',
             'NTPCustomBackgroundEnabled',
             'NativeWindowOcclusionEnabled',
             'NearbyShareAllowed',
             'PaymentMethodQueryEnabled',
             'PdfAnnotationsEnabled',
             'PhoneHubAllowed',
             'PhoneHubNotificationsAllowed',
             'PhoneHubTaskContinuationAllowed',
             'PolicyAtomicGroupsEnabled',
             'PrimaryMouseButtonSwitch',
             'PromotionalTabsEnabled',
             'PromptForDownloadLocation',
             'QuicAllowed',
             'RendererCodeIntegrityEnabled',
             'RequireOnlineRevocationChecksForLocalAnchors',
             'RoamingProfileSupportEnabled',
             'SSLErrorOverrideAllowed',
             'SafeBrowsingForTrustedSourcesEnabled',
             'SavingBrowserHistoryDisabled',
             'ScreenCaptureAllowed',
             'ScrollToTextFragmentEnabled',
             'SearchSuggestEnabled',
             'SecondaryGoogleAccountSigninAllowed',
             'SharedArrayBufferUnrestrictedAccessAllowed',
             'SharedClipboardEnabled',
             'ShowAppsShortcutInBookmarkBar',
             'ShowFullUrlsInAddressBar',
             'ShowLogoutButtonInTray',
             'SignedHTTPExchangeEnabled',
             'SigninAllowed',
             'SigninInterceptionEnabled',
             'SitePerProcess',
             'SmartLockSigninAllowed',
             'SmsMessagesAllowed',
             'SpellCheckServiceEnabled',
             'SpellcheckEnabled',
             'StartupBrowserWindowLaunchSuppressed',
             'StricterMixedContentTreatmentEnabled',
             'SuggestLogoutAfterClosingLastWindow',
             'SuppressDifferentOriginSubframeDialogs',
             'SuppressUnsupportedOSWarning',
             'SyncDisabled',
             'TargetBlankImpliesNoOpener',
             'TaskManagerEndProcessEnabled',
             'ThirdPartyBlockingEnabled',
             'TouchVirtualKeyboardEnabled',
             'TranslateEnabled',
             'TripleDESEnabled',
             'UnifiedDesktopEnabledByDefault',
             'UrlKeyedAnonymizedDataCollectionEnabled',
             'UserAgentClientHintsEnabled',
             'UserFeedbackAllowed',
             'VideoCaptureAllowed',
             'VmManagementCliAllowed',
             'VpnConfigAllowed',
             'WPADQuickCheckEnabled',
             'WebRtcAllowLegacyTLSProtocols',
             'WebRtcEventLogCollectionAllowed',
             'WifiSyncAndroidAllowed',
             'WindowOcclusionEnabled']
    if name in dict_entries:
        return json.loads(get_string(e.data))
    elif e.type == misc.REG_DWORD and name in bools:
        return e.data == 1
    return e.data

def assign_entry(policies, e):
    if e.valuename.isnumeric():
        name = e.keyname.split('\\')[-1]
        if name not in policies:
            policies[name] = []
        policies[name].append(parse_entry_data(name, e))
    else:
        name = e.valuename
        policies[name] = parse_entry_data(name, e)

def convert_pol_to_json(section, entries):
    managed = {}
    recommended = {}
    recommended_section = '\\'.join([section, 'Recommended'])
    for e in entries:
        if '**delvals.' in e.valuename:
            continue
        if e.keyname.startswith(recommended_section):
            assign_entry(recommended, e)
        elif e.keyname.startswith(section):
            assign_entry(managed, e)
    return managed, recommended

class gp_chromium_ext(gp_pol_ext, gp_file_applier):
    managed_policies_path = '/etc/chromium/policies/managed'
    recommended_policies_path = '/etc/chromium/policies/recommended'

    def __str__(self):
        return 'Google/Chromium'

    def process_group_policy(self, deleted_gpo_list, changed_gpo_list,
                             policy_dir=None):
        if policy_dir is not None:
            self.recommended_policies_path = os.path.join(policy_dir,
                                                            'recommended')
            self.managed_policies_path = os.path.join(policy_dir, 'managed')
        # Create the policy directories if necessary
        if not os.path.exists(self.recommended_policies_path):
            os.makedirs(self.recommended_policies_path, mode=0o755,
                        exist_ok=True)
        if not os.path.exists(self.managed_policies_path):
            os.makedirs(self.managed_policies_path, mode=0o755,
                        exist_ok=True)
        for guid, settings in deleted_gpo_list:
            if str(self) in settings:
                for attribute, policies in settings[str(self)].items():
                    try:
                        json.loads(policies)
                    except json.decoder.JSONDecodeError:
                        self.unapply(guid, attribute, policies)
                    else:
                        # Policies were previously stored all in one file, but
                        # the Chromium documentation says this is not
                        # necessary. Unapply the old policy file if json was
                        # stored in the cache (now we store a hash and file
                        # names instead).
                        if attribute == 'recommended':
                            fname = os.path.join(self.recommended_policies_path,
                                                 'policies.json')
                        elif attribute == 'managed':
                            fname = os.path.join(self.managed_policies_path,
                                                 'policies.json')
                        self.unapply(guid, attribute, fname)

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                section = 'Software\\Policies\\Google\\Chrome'
                pol_file = 'MACHINE/Registry.pol'
                path = os.path.join(gpo.file_sys_path, pol_file)
                pol_conf = self.parse(path)
                if not pol_conf:
                    continue

                managed, recommended = convert_pol_to_json(section,
                                                           pol_conf.entries)
                def applier_func(policies, location):
                    try:
                        with NamedTemporaryFile(mode='w+', prefix='gp_',
                                                delete=False,
                                                dir=location,
                                                suffix='.json') as f:
                            json.dump(policies, f)
                            os.chmod(f.name, 0o644)
                            log.debug('Wrote Chromium preferences', policies)
                            return [f.name]
                    except PermissionError:
                        log.debug('Failed to write Chromium preferences',
                                  policies)
                value_hash = self.generate_value_hash(json.dumps(managed))
                self.apply(gpo.name, 'managed', value_hash, applier_func,
                           managed, self.managed_policies_path)
                value_hash = self.generate_value_hash(json.dumps(recommended))
                self.apply(gpo.name, 'recommended', value_hash, applier_func,
                           recommended, self.recommended_policies_path)

    def rsop(self, gpo):
        output = {}
        pol_file = 'MACHINE/Registry.pol'
        section = 'Software\\Policies\\Google\\Chrome'
        if gpo.file_sys_path:
            path = os.path.join(gpo.file_sys_path, pol_file)
            pol_conf = self.parse(path)
            if not pol_conf:
                return output
            for e in pol_conf.entries:
                if e.keyname.startswith(section):
                    output['%s\\%s' % (e.keyname, e.valuename)] = e.data
        return output

class gp_chrome_ext(gp_chromium_ext):
    managed_policies_path = '/etc/opt/chrome/policies/managed'
    recommended_policies_path = '/etc/opt/chrome/policies/recommended'

    def __str__(self):
        return 'Google/Chrome'



# Blackbox tests for the "net ads ... --json" commands
# Copyright (C) 2018 Intra2net AG
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import re

import samba.tests
from samba.common import get_string

COMMAND         = "bin/net ads"
# extract keys from non-json version
PLAIN_KEY_REGEX = re.compile ("^([^ \t:][^:]*):")

class BaseWrapper (object):
    """
    Guard the base so it doesn't inherit from TestCase. This prevents it from
    being run by unittest directly.
    """

    class NetAdsJSONTests_Base(samba.tests.BlackboxTestCase):
        """Blackbox tests for JSON output of the net ads suite of commands."""
        subcmd = None

        def test_json_wellformed (self):
            """The output of ``--json`` commands must parse as JSON."""
            argv = "%s %s --json" % (COMMAND, self.subcmd)
            try:
                out = self.check_output(argv)
                json.loads (get_string(out))
            except samba.tests.BlackboxProcessError as e:
                self.fail("Error calling [%s]: %s" % (argv, e))

        def test_json_matching_entries (self):
            """
            The ``--json`` variants must contain the same keys as their
            respective plain counterpart.

            Does not check nested dictionaries (e. g. the ``Flags`` value of
            ``net ads lookup``..
            """
            argv = "%s %s" % (COMMAND, self.subcmd)
            try:
                out_plain = get_string(self.check_output(argv))
            except samba.tests.BlackboxProcessError as e:
                self.fail("Error calling [%s]: %s" % (argv, e))

            argv = "%s %s --json" % (COMMAND, self.subcmd)
            try:
                out_jsobj = self.check_output(argv)
            except samba.tests.BlackboxProcessError as e:
                self.fail("Error calling [%s]: %s" % (argv, e))

            parsed = json.loads (get_string(out_jsobj))

            for key in [ re.match (PLAIN_KEY_REGEX, line).group(1)
                         for line in out_plain.split ("\n")
                            if line != "" and line [0] not in " \t:" ]:
                self.assertTrue (parsed.get (key) is not None)
                del parsed [key]

            self.assertTrue (len (parsed) == 0) # tolerate no leftovers

class NetAdsJSONInfoTests(BaseWrapper.NetAdsJSONTests_Base):
    subcmd = "info"

class NetAdsJSONlookupTests(BaseWrapper.NetAdsJSONTests_Base):
    subcmd = "lookup"



# Unix SMB/CIFS implementation. Tests for dsdb
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2023
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.dsdb."""

from samba.credentials import Credentials
from samba.samdb import SamDB
from samba.auth import system_session
from samba.param import LoadParm
from samba import dsdb, functional_level
import ldb


from samba.tests.samba_tool.base import SambaToolCmdTest
import os
import shutil
import tempfile

class SambaFLStartUpTests(SambaToolCmdTest):
    """Test the samba binary sets the DC FL on startup for RW DCs"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.classtempdir = tempfile.mkdtemp()
        cls.tempsambadir = os.path.join(cls.classtempdir, "samba")

        command = (
                "samba-tool " +
                "domain provision " +
                "--realm=foo.example.com " +
                "--domain=FOO " +
                ("--targetdir=%s " % cls.tempsambadir) +
                "--use-ntvfs"
        )

        (result, out, err) = cls.run_command(command)
        if (result != 0):
            raise AssertionError

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        shutil.rmtree(cls.tempsambadir)

    def setUp(self):
        super().setUp()
        path = os.path.join(self.tempsambadir, "etc/smb.conf")
        self.lp = LoadParm(filename_for_non_global_lp=path)
        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.session = system_session()
        self.samdb = SamDB(session_info=self.session,
                           credentials=self.creds,
                           lp=self.lp)


    def test_initial_db_fl_state(self):
        server_dn = self.samdb.get_dsServiceName()
        res = self.samdb.search(base=server_dn,
                                 scope=ldb.SCOPE_BASE,
                                 attrs=["msDS-Behavior-Version"])
        # This confirms the domain is in FL 2008 R2 by default, this is
        # important to verify the original state
        self.assertEqual(int(res[0]["msDS-Behavior-Version"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2008_R2)

    def test_initial_rootdse_domain_fl_state(self):
        res = self.samdb.search(base="",
                                scope=ldb.SCOPE_BASE,
                                attrs=["domainControllerFunctionality"])
        self.assertEqual(int(res[0]["domainControllerFunctionality"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2008_R2)

    def test_initial_rootdse_dc_fl_state(self):
        res = self.samdb.search(base="",
                                scope=ldb.SCOPE_BASE,
                                attrs=["domainFunctionality"])
        self.assertEqual(int(res[0]["domainFunctionality"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2008_R2)

    def test_initial_lp_fl_state(self):
        lp_fl = self.lp.get("ad dc functional level")
        # This confirms the domain is in FL 2008 R2 by default, this is
        # important to verify the original state
        self.assertEqual(lp_fl, "2008_R2")

    def test_initial_lp_fl_state_mapped(self):
        # Confirm the same via the dc_level_from_lp wrapper
        self.assertEqual(functional_level.dc_level_from_lp(self.lp),
                         dsdb.DS_DOMAIN_FUNCTION_2008_R2)

    def fixup_fl(self, dn, fl):
        msg = ldb.Message()
        msg.dn = dn
        msg["msDS-Behavior-Version"] = (
            ldb.MessageElement(str(fl),
                               ldb.FLAG_MOD_REPLACE,
                               "msDS-Behavior-Version"))
        self.samdb.modify(msg)

    def test_change_db_dc_fl(self):
        server_dn = ldb.Dn(self.samdb, self.samdb.get_dsServiceName())
        msg = ldb.Message()
        msg.dn = server_dn
        msg["msDS-Behavior-Version"] = (
                ldb.MessageElement(str(dsdb.DS_DOMAIN_FUNCTION_2012_R2),
                                   ldb.FLAG_MOD_REPLACE,
                                   "msDS-Behavior-Version"))
        self.samdb.modify(msg)
        self.addCleanup(self.fixup_fl, msg.dn, dsdb.DS_DOMAIN_FUNCTION_2008_R2)

        samdb2 = SamDB(session_info=self.session,
                       credentials=self.creds,
                       lp=self.lp)

        # Check that the DB set to 2012_R2 has got as far as the rootDSE handler on a new connection
        res = samdb2.search(base="",
                            scope=ldb.SCOPE_BASE,
                            attrs=["domainControllerFunctionality"])
        self.assertEqual(int(res[0]["domainControllerFunctionality"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2012_R2)

    def test_incorrect_db_dc_fl(self):
        server_dn = ldb.Dn(self.samdb, self.samdb.get_dsServiceName())
        self.addCleanup(self.fixup_fl, server_dn, dsdb.DS_DOMAIN_FUNCTION_2008_R2)

        old_lp_fl = self.lp.get("ad dc functional level")
        self.lp.set("ad dc functional level",
                    "2016")
        self.addCleanup(self.lp.set, "ad dc functional level", old_lp_fl)

        dsdb.check_and_update_fl(self.samdb, self.lp)

        # Check this has been set to 2016 per the smb.conf setting
        res = self.samdb.search(base="",
                                 scope=ldb.SCOPE_BASE,
                                 attrs=["domainControllerFunctionality"])
        self.assertEqual(int(res[0]["domainControllerFunctionality"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2016)

        samdb3 = SamDB(session_info=self.session,
                       credentials=self.creds,
                       lp=self.lp)

        # Check this is still set on re-read (not just the opaque)
        res = samdb3.search(base="",
                            scope=ldb.SCOPE_BASE,
                            attrs=["domainControllerFunctionality"])
        self.assertEqual(int(res[0]["domainControllerFunctionality"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2016)

        res = self.samdb.search(base=server_dn,
                                 scope=ldb.SCOPE_BASE,
                                 attrs=["msDS-Behavior-Version"])
        self.assertEqual(int(res[0]["msDS-Behavior-Version"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2016)

        self.assertEqual(functional_level.dc_level_from_lp(self.lp),
                         dsdb.DS_DOMAIN_FUNCTION_2016)
        self.assertEqual(self.lp.get("ad dc functional level"),
                         "2016")

if __name__ == "__main__":
    import unittest
    unittest.main()



# Tests basic behaviour when NTLM is disabled
#
# Copyright (C) Catalyst IT Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from samba.tests import TestCase
import os

import samba
from samba.credentials import Credentials, DONT_USE_KERBEROS, MUST_USE_KERBEROS

from samba import NTSTATUSError, ntstatus
import ctypes

from samba.dcerpc import srvsvc, samr, lsa

"""
Tests behaviour when NTLM is disabled
"""


class NtlmDisabledTests(TestCase):

    def setUp(self):
        super().setUp()

        self.lp          = self.get_loadparm()
        self.server      = os.getenv("SERVER")

        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.creds.set_username(os.getenv("USERNAME"))
        self.creds.set_domain(self.server)
        self.creds.set_password(os.getenv("PASSWORD"))
        self.creds.set_kerberos_state(DONT_USE_KERBEROS)

    def test_ntlm_connection(self):
        try:
            conn = srvsvc.srvsvc("ncacn_np:%s[smb2,ntlm]" % self.server, self.lp, self.creds)

            self.assertIsNotNone(conn)
        except NTSTATUSError as e:
            # NTLM might be blocked on this server
            enum = ctypes.c_uint32(e.args[0]).value
            if enum == ntstatus.NT_STATUS_NTLM_BLOCKED:
                self.fail("NTLM is disabled on this server")
            else:
                raise

    def test_samr_change_password(self):
        self.creds.set_kerberos_state(MUST_USE_KERBEROS)
        conn = samr.samr("ncacn_np:%s[krb5,seal,smb2]" % os.getenv("SERVER"))

        # we want to check whether this gets rejected outright because NTLM is
        # disabled, so we don't actually need to encrypt a valid password here
        server = lsa.String()
        server.string = self.server
        username = lsa.String()
        username.string = os.getenv("USERNAME")

        try:
            conn.ChangePasswordUser2(server, username, None, None, True, None, None)
        except NTSTATUSError as e:
            # changing passwords should be rejected when NTLM is disabled
            enum = ctypes.c_uint32(e.args[0]).value
            if enum == ntstatus.NT_STATUS_NTLM_BLOCKED:
                self.fail("NTLM is disabled on this server")
            elif enum == ntstatus.NT_STATUS_WRONG_PASSWORD:
                # expected error case when NTLM is enabled
                pass
            else:
                raise



# Unix SMB/CIFS implementation.
# Copyright (C) David Mulder <dmulder@samba.org> 2021
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import re
from samba.tests.samba_tool.base import SambaToolCmdTest
from samba.param import LoadParm
from samba.netcmd.common import netcmd_dnsname

class JoinMemberCmdTestCase(SambaToolCmdTest):
    """Test for samba-tool domain join subcommand"""

    def test_join_member(self):
        """Run a domain member join, and check that dns is updated"""
        smb_conf = os.environ["SERVERCONFFILE"]
        zone = os.environ["REALM"].lower()
        lp = LoadParm()
        lp.load(smb_conf)
        dnsname = netcmd_dnsname(lp)
        # Fetch the existing dns A records
        (result, out, err) = self.runsubcmd("dns", "query",
                                    os.environ["DC_SERVER"],
                                    zone, dnsname, 'A',
                                    "-s", smb_conf,
                                    "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                  os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, 'Failed to find the record')

        existing_records = re.findall(r'A:\s+(\d+\.\d+\.\d+\.\d+)\s', out)

        # Remove the existing records
        for record in existing_records:
            (result, out, err) = self.runsubcmd("dns", "delete",
                                        os.environ["DC_SERVER"],
                                        zone, dnsname, 'A', record,
                                        "-s", smb_conf,
                                        "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                      os.environ["DC_PASSWORD"]))
            self.assertCmdSuccess(result, out, err, 'Failed to remove record')

        # Perform the s3 member join (net ads join)
        (result, out, err) = self.runsubcmd("domain", "join",
                                    os.environ["REALM"], "member",
                                    "-s", smb_conf,
                                    "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                  os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, 'Failed to join member')

        # Ensure the dns A record was created
        (result, out, err) = self.runsubcmd("dns", "query",
                                    os.environ["DC_SERVER"],
                                    zone, dnsname, 'A',
                                    "-s", smb_conf,
                                    "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                  os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err,
                              'Failed to find dns host records for %s' % dnsname)



# Unix SMB/CIFS implementation. Tests for dsdb
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2021
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.dsdb."""

from samba.tests import TestCase, DynamicTestCase
from samba.dsdb import user_account_control_flag_bit_to_string
import samba


@DynamicTestCase
class DsdbFlagTests(TestCase):

    @classmethod
    def setUpDynamicTestCases(cls):

        for x in dir(samba.dsdb):
            if x.startswith("UF_"):
                cls.generate_dynamic_test("test",
                                          x,
                                          x,
                                          getattr(samba.dsdb, x))


    def _test_with_args(self, uf_string, uf_bit):
        self.assertEqual(user_account_control_flag_bit_to_string(uf_bit),
                         uf_string)


    def test_not_a_flag(self):
        self.assertRaises(KeyError,
                          user_account_control_flag_bit_to_string,
                          0xabcdef)

    def test_too_long(self):
        self.assertRaises(OverflowError,
                          user_account_control_flag_bit_to_string,
                          0xabcdefffff)

    def test_way_too_long(self):
        self.assertRaises(OverflowError,
                          user_account_control_flag_bit_to_string,
                          0xabcdeffffffffffff)



# Unix SMB/CIFS implementation.
# Copyright (C) Douglas Bagnall <douglas.bagnall@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import tarfile
from samba import safe_tarfile

import os
from samba.tests import TestCaseInTempDir


def filterer(prefix):
    def f(info):
        info.name = prefix + info.name
        return info
    return f


class SafeTarFileTestCase(TestCaseInTempDir):

    def test_dots(self):
        filename = os.path.join(self.tempdir, 'x')
        tarname = os.path.join(self.tempdir, 'tar.tar')
        f = open(filename, 'w')
        f.write('x')
        f.close()

        tf = tarfile.open(tarname, 'w')
        tf.add(filename, filter=filterer('../../'))
        tf.close()

        stf = safe_tarfile.open(tarname)

        # If we have data_filter, we have a patched python to address
        # CVE-2007-4559.
        if hasattr(tarfile, "data_filter"):
            self.assertRaises((tarfile.OutsideDestinationError,
                               NotADirectoryError),
                              stf.extractall,
                              tarname)
        else:
            self.assertRaises(tarfile.ExtractError,
                              stf.extractall,
                              tarname)
        self.rm_files('x', 'tar.tar')

    def test_slash(self):
        filename = os.path.join(self.tempdir, 'x')
        tarname = os.path.join(self.tempdir, 'tar.tar')
        f = open(filename, 'w')
        f.write('x')
        f.close()

        tf = tarfile.open(tarname, 'w')
        tf.add(filename, filter=filterer('/'))
        tf.close()

        stf = safe_tarfile.open(tarname)

        # If we have data_filter, we have a patched python to address
        # CVE-2007-4559.
        if hasattr(tarfile, "data_filter"):
            self.assertRaises(NotADirectoryError,
                              stf.extractall,
                              tarname)
        else:
            self.assertRaises(tarfile.ExtractError,
                              stf.extractall,
                              tarname)

        self.rm_files('x', 'tar.tar')



# Unix SMB/CIFS implementation.
#
# Query class for the ORM to the Ldb database.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re

from .exceptions import NotFound, MultipleObjectsReturned
from .registry import MODELS

RE_SPLIT_CAMELCASE = re.compile(r"[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))")


class Query:
    """Simple Query class used by the `Model.query` method."""

    def __init__(self, model, samdb, result, polymorphic):
        self.model = model
        self.samdb = samdb
        self.result = result
        self.count = result.count
        self.name = " ".join(RE_SPLIT_CAMELCASE.findall(model.__name__)).lower()
        self.polymorphic = polymorphic

    def __iter__(self):
        """Loop over Query class yields Model instances."""
        for message in self.result:
            yield self._from_message(message)

    def __repr__(self):
        """Provide repr method that provides more useful output in the shell."""
        return f"<Query {list(self)}>"

    def _from_message(self, message):
        """Returns the model class to use to construct instances.

        If polymorphic query is enabled it will use the last item from
        the objectClass list.

        Otherwise, it will use the model from the queryset.
        """
        if self.polymorphic:
            object_class = str(message["objectClass"][-1])
            model = MODELS.get(object_class, self.model)
        else:
            model = self.model

        return model._from_message(self.samdb, message)

    def first(self):
        """Returns the first item in the Query or None for no results."""
        if self.count:
            return self._from_message(self.result[0])

    def last(self):
        """Returns the last item in the Query or None for no results."""
        if self.count:
            return self._from_message(self.result[-1])

    def get(self):
        """Returns one item or None if no results were found.

        :returns: Model instance or None if not found.
        :raises MultipleObjectsReturned: if more than one results were returned
        """
        if self.count > 1:
            raise MultipleObjectsReturned(
                f"More than one {self.name} objects returned (got {self.count}).")
        elif self.count:
            return self._from_message(self.result[0])

    def one(self):
        """Must return EXACTLY one item or raise an exception.

        :returns: Model instance
        :raises NotFound: if no results were returned
        :raises MultipleObjectsReturned: if more than one results were returned
        """
        if self.count < 1:
            raise NotFound(
                f"{self.name.capitalize()} matching query not found")
        elif self.count > 1:
            raise MultipleObjectsReturned(
                f"More than one {self.name} objects returned (got {self.count}).")
        else:
            return self._from_message(self.result[0])



# vgp_motd_ext samba gpo policy
# Copyright (C) David Mulder <dmulder@suse.com> 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from samba.gp.gpclass import gp_xml_ext, gp_misc_applier

class vgp_motd_ext(gp_xml_ext, gp_misc_applier):
    def unapply(self, guid, motd, attribute, value):
        if attribute != 'motd':
            raise ValueError('"%s" is not a message attribute' % attribute)
        msg = value
        data = self.parse_value(value)
        if os.path.exists(motd):
            with open(motd, 'r') as f:
                current = f.read()
        else:
            current = ''
        # Only overwrite the msg if it hasn't been modified. It may have been
        # modified by another GPO.
        if 'new_val' not in data or current.strip() == data['new_val'].strip():
            msg = data['old_val']
            with open(motd, 'w') as w:
                if msg:
                    w.write(msg)
                else:
                    w.truncate()
        self.cache_remove_attribute(guid, attribute)

    def apply(self, guid, motd, text):
        if os.path.exists(motd):
            with open(motd, 'r') as f:
                current = f.read()
        else:
            current = ''
        if current != text.text:
            with open(motd, 'w') as w:
                w.write(text.text)
            data = self.generate_value(old_val=current, new_val=text.text)
            self.cache_add_attribute(guid, 'motd', data)

    def __str__(self):
        return 'Unix Settings/Message of the Day'

    def process_group_policy(self, deleted_gpo_list, changed_gpo_list,
                             motd='/etc/motd'):
        for guid, settings in deleted_gpo_list:
            if str(self) in settings:
                for attribute, msg in settings[str(self)].items():
                    self.unapply(guid, motd, attribute, msg)

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                xml = 'MACHINE/VGP/VTLA/Unix/MOTD/manifest.xml'
                path = os.path.join(gpo.file_sys_path, xml)
                xml_conf = self.parse(path)
                if not xml_conf:
                    continue
                policy = xml_conf.find('policysetting')
                data = policy.find('data')
                text = data.find('text')
                self.apply(gpo.name, motd, text)

    def rsop(self, gpo):
        output = {}
        if gpo.file_sys_path:
            xml = 'MACHINE/VGP/VTLA/Unix/MOTD/manifest.xml'
            path = os.path.join(gpo.file_sys_path, xml)
            xml_conf = self.parse(path)
            if not xml_conf:
                return output
            policy = xml_conf.find('policysetting')
            data = policy.find('data')
            filename = data.find('filename')
            text = data.find('text')
            mfile = os.path.join('/etc', filename.text)
            output[mfile] = text.text
        return output



# Unix SMB/CIFS implementation.
# Copyright (C) Volker Lendecke <vl@samba.org> 2023
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from samba import Ldb,tests
from samba.samba3 import param as s3param
from samba import credentials
import os

class LdapWhoami(tests.TestCase):
    def test_ldap_whoami(self):
        lp = s3param.get_context()
        lp.load(os.getenv("SERVERCONFFILE"))

        domain=os.getenv("DOMAIN")
        username=os.getenv("DC_USERNAME")

        creds = credentials.Credentials()
        creds.guess(lp)
        creds.set_domain(domain)
        creds.set_username(username)
        creds.set_password(os.getenv("DC_PASSWORD"))

        l=Ldb(f'ldap://{os.getenv("DC_SERVER_IP")}/', credentials=creds, lp=lp)
        w=l.whoami()
        self.assertEqual(w,f'u:{domain}\\{username}')



# Unix SMB/CIFS implementation.
#
# Copyright (C) Catalyst.Net Ltd. 2022
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import subprocess
import os
from samba.tests import TestCaseInTempDir
from pprint import pprint

HERE = os.path.dirname(__file__)
S4_SERVER = os.path.join(HERE, '../../../../bin/test_s4_logging')
S3_SERVER = os.path.join(HERE, '../../../../bin/test_s3_logging')

CLASS_LIST = ["all", "tdb", "printdrivers", "lanman", "smb",
              "rpc_parse", "rpc_srv", "rpc_cli", "passdb", "sam", "auth",
              "winbind", "vfs", "idmap", "quota", "acls", "locking", "msdfs",
              "dmapi", "registry", "scavenger", "dns", "ldb", "tevent",
              "auth_audit", "auth_json_audit", "kerberos", "drs_repl",
              "smb2", "smb2_credits", "dsdb_audit", "dsdb_json_audit",
              "dsdb_password_audit", "dsdb_password_json_audit",
              "dsdb_transaction_audit", "dsdb_transaction_json_audit",
              "dsdb_group_audit", "dsdb_group_json_audit", "ldapsrv"]


CLASS_CODES = {k: i for i, k in enumerate(CLASS_LIST)}


class S4LoggingTests(TestCaseInTempDir):
    server = S4_SERVER
    def _write_smb_conf(self,
                        default_level=2,
                        default_file="default",
                        mapping=()):
        self.smbconf = os.path.join(self.tempdir, "smb.conf")

        with open(self.smbconf, "w") as f:
            f.write('[global]\n')
            if default_file is not None:
                dest = os.path.join(self.tempdir,
                                    default_file)
                f.write(f"    log file = {dest}\n")

            f.write("    log level = ")
            if default_level:
                f.write(f"{default_level}")

            for dbg_class, log_level, log_file in mapping:
                f.write(' ')
                f.write(dbg_class)
                if log_level is not None:
                    f.write(f':{log_level}')
                if log_file is not None:
                    dest = os.path.join(self.tempdir,
                                        log_file)

                    f.write(f'@{dest}')
            f.write('\n')
        self.addCleanup(os.unlink, self.smbconf)

    def _extract_log_level_line(self, new_level=2):
        # extricate the 'log level' line from the smb.conf, returning
        # the value, and replacing the log level line with something
        # innocuous.
        smbconf2 = self.smbconf + 'new'
        with open(self.smbconf) as f:
            with open(smbconf2, 'w') as f2:
                for line in f:
                    if 'log level' in line:
                        debug_arg = line.split('=', 1)[1].strip()
                        if new_level is not None:
                            f2.write(f'    log level = {new_level}\n')
                    else:
                        f2.write(line)
        os.replace(smbconf2, self.smbconf)
        return debug_arg

    def _get_expected_strings(self, mapping,
                              level_filter,
                              default_file='default',
                              file_filter=None):
        default = os.path.join(self.tempdir, default_file)
        expected = {default: []}
        # this kind of thing:
        # "  logging for 'dns' [21], at level 4"
        for dbg_class, log_level, log_file in mapping:
            if log_file is None:
                log_file = default_file

            f = os.path.join(self.tempdir, log_file)
            expected.setdefault(f, [])
            if log_level < level_filter:
                continue
            if file_filter not in (None, log_file):
                continue
            s = (f"  logging for '{dbg_class}' [{CLASS_CODES[dbg_class]}], "
                 f"at level {level_filter}")
            expected[f].append(s)

        return expected

    def _run_s4_logger(self, log_level, *extra_args):
        cmd = [self.server,
               '-s', self.smbconf,
               '-L', str(log_level),
               *extra_args]

        p = subprocess.run(cmd,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
        self.assertEqual(p.returncode, 0,
                         f"'{' '.join(cmd)}' failed ({p.returncode})")

        return p.stdout.decode(), p.stderr.decode()

    def assert_string_contains(self, string, expected_lines,
                               filename=None):
        expected_lines = set(expected_lines)
        string_lines = set(string.split('\n'))
        present_lines = string_lines & expected_lines
        if present_lines != expected_lines:
            if filename:
                print(filename)
            print("expected %d lines, found %d" %
                  (len(expected_lines), len(present_lines)))
            print("missing lines:")
            pprint(expected_lines - present_lines)
            raise AssertionError("missing lines")

    def assert_file_contains(self, filename, expected_lines):
        with open(filename) as f:
            string = f.read()
        self.assert_string_contains(string, expected_lines, filename)

    def assert_n_known_lines_string(self, string, n):
        count = string.count("logging for '")
        if count != n:
            raise AssertionError(
                f"string has {count} lines, expected {n}")

    def assert_n_known_lines(self, filename, n):
        with open(filename) as f:
            string = f.read()
        count = string.count("  logging for '")
        if count != n:
            raise AssertionError(
                f"{filename} has {count} lines, expected {n}")

    def assert_unlink_expected_strings(self, expected_strings):
        for k, v in expected_strings.items():
            if not os.path.exists(k):
                self.fail(f"{k} does not exist")
            self.assert_file_contains(k, v)
            self.assert_n_known_lines(k, len(v))
            os.unlink(k)

    def test_each_to_its_own(self):
        level = 4
        mapping = [(x, level, x) for x in CLASS_LIST]
        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_all_to_one(self):
        level = 4
        dest = 'everything'
        mapping = [(x, level, dest) for x in CLASS_LIST]
        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_bifurcate(self):
        level = 4
        dests = ['even', 'odd']
        mapping = [(x, level + 1, dests[i & 1])
                   for i, x in enumerate(CLASS_LIST)]
        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_bifurcate_level_out_of_range(self):
        # nothing will be logged, because we're logging at a too high
        # level.
        level = 4
        dests = ['even', 'odd']
        mapping = [(x, level - 1, dests[i & 1])
                   for i, x in enumerate(CLASS_LIST)]
        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_bifurcate_misc_log_level(self):
        # We are sending even numbers to default and odd numbers to
        # 'odd', at various levels, depending on mod 3. Like this:
        #
        # log level = 2 all:5 \
        #               tdb:4@odd \
        #               printdrivers:3 \
        #               lanman:5@odd \
        #               smb:4 \
        #               rpc_parse:3@odd \
        #               rpc_srv:5 ...
        #
        # Therefore, 'default' should get classes that are (0 or 4) % 6
        # and 'odd' should get classes that are (1 or 3) % 6.

        level = 4
        dests = [None, 'odd']
        mapping = []
        for i, x in enumerate(CLASS_LIST):
            parity = i & 1
            log_level = level + 1 - (i % 3)
            mapping.append((x, log_level, dests[parity]))

        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_all_different_ways_cmdline_d(self):
        level = 4
        dests = [None, 'a', 'b', 'c']
        mapping = []
        seed = 123
        for i, x in enumerate(CLASS_LIST):
            d = seed & 3
            seed = seed * 17 + 1
            log_level = seed % 10
            seed &= 0xff
            mapping.append((x, log_level, dests[d]))

        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        debug_arg = self._extract_log_level_line(26)

        stdout, stderr = self._run_s4_logger(level, '-d', debug_arg)
        self.assert_unlink_expected_strings(expected_strings)

    def test_all_different_ways_cmdline_d_interactive(self):
        level = 4
        dests = [None, 'a', 'b', 'c']
        mapping = []
        seed = 1234
        for i, x in enumerate(CLASS_LIST):
            d = seed & 3
            seed = seed * 13 + 1
            log_level = seed % 10
            seed &= 0xff
            mapping.append((x, log_level, dests[d]))

        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        debug_arg = self._extract_log_level_line(None)
        stdout, stderr = self._run_s4_logger(level, '-d', debug_arg, '-i')
        expected_lines = []
        for v in expected_strings.values():
            # stderr doesn't end up with leading '  '
            expected_lines.extend([x.strip() for x in v])

        self.assert_string_contains(stderr, expected_lines)
        self.assert_n_known_lines_string(stderr, len(expected_lines))

    def test_only_some_level_0(self):
        # running the logger with -L 0 makes the log messages run at
        # level 0 (i.e DBG_ERR), so we always see them in default,
        # even though smb.conf doesn't ask.
        mapping = [(x, 3, ['default', 'bees']['b' in x])
                   for x in CLASS_LIST]
        expected_strings = self._get_expected_strings(mapping, 0)
        self._write_smb_conf(mapping=[x for x in mapping if x[2] == 'bees'])
        stdout, stderr = self._run_s4_logger(0)
        self.assert_unlink_expected_strings(expected_strings)

    def test_only_some_level_3(self):
        # here, we're expecting the unmentioned non-b classes to just
        # disappear.
        level = 3
        mapping = [(x, level, 'bees') for x in CLASS_LIST if 'b' in x]
        expected_strings = self._get_expected_strings(mapping, level)
        self._write_smb_conf(mapping=[x for x in mapping if x[2] == 'bees'])
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_none(self):
        level = 4
        mapping = []
        expected_strings = self._get_expected_strings(mapping, level)
        self._write_smb_conf(mapping=mapping)
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_none_high_default(self):
        # We set the default level to 5 and do nothing else special,
        # which means we need a different mapping for the smb.conf
        # than the expected strings.
        level = 4
        mapping = [(x, 5, 'default') for x in CLASS_LIST]
        expected_strings = self._get_expected_strings(mapping, level)
        # note the empty mapping in smb.conf
        self._write_smb_conf(mapping=[], default_level=5)
        stdout, stderr = self._run_s4_logger(level)
        self.assert_unlink_expected_strings(expected_strings)

    def test_none_high_cmdline_d(self):
        # We set the default level to 2, but run the 'server' with -d 10.
        level = 4
        mapping = [(x, 10, 'default') for x in CLASS_LIST]
        expected_strings = self._get_expected_strings(mapping, level)
        # note the empty mapping in smb.conf
        self._write_smb_conf(mapping=[])
        stdout, stderr = self._run_s4_logger(level, '-d', '10')
        self.assert_unlink_expected_strings(expected_strings)

    def test_interactive_high_default_simple(self):
        # running with -i should send everything to stderr.
        level = 4
        mapping = [(x, 5, 'default') for x in CLASS_LIST]
        expected_strings = self._get_expected_strings(mapping, level)
        self._write_smb_conf(mapping=[], default_level=5)
        stdout, stderr = self._run_s4_logger(level, '-i')
        expected_lines = []
        for v in expected_strings.values():
            # stderr doesn't end up with leading '  '
            expected_lines.extend([x.strip() for x in v])

        self.assert_string_contains(stderr, expected_lines)

    def test_interactive_complex_smb_conf(self):
        # running with -i should send everything to stderr. The
        # smb.conf will set the levels, but the target files are
        # overridden.
        # (this is the test_bifurcate_misc_log_level() smb.conf).
        level = 4
        dests = [None, 'odd']
        mapping = []
        for i, x in enumerate(CLASS_LIST):
            parity = i & 1
            log_level = level + 1 - (i % 3)
            mapping.append((x, log_level, dests[parity]))

        expected_strings = self._get_expected_strings(mapping, level)

        self._write_smb_conf(mapping=mapping)
        stdout, stderr = self._run_s4_logger(level, '-i')
        expected_lines = []
        for v in expected_strings.values():
            # stderr doesn't end up with leading '  '
            expected_lines.extend([x.strip() for x in v])

        self.assert_string_contains(stderr, expected_lines)


class S3LoggingTests(S4LoggingTests):
    server = S3_SERVER
    # These tests were developed for testing the test_logger when
    # linked against CMDLINE_S4 (see lib/util/wscript_build), but can
    # also run when linked against CMDLINE_S3.



# test generated python code from pidl
# Copyright (C) Andrew Tridgell August 2010
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# to run this test, use one of these:
#
#    python -m unittest samba.tests.dcerpc.rpc_talloc
#
# or if you have trial installed (from twisted), use
#
#    trial samba.tests.dcerpc.rpc_talloc

"""Tests for the talloc handling in the generated Python DCE/RPC bindings."""

import sys

sys.path.insert(0, "bin/python")

import samba
import samba.tests
from samba.dcerpc import drsuapi
import talloc

talloc.enable_null_tracking()


class TallocTests(samba.tests.TestCase):
    """test talloc behaviour of pidl generated python code"""

    def check_blocks(self, object, num_expected):
        """check that the number of allocated blocks is correct"""
        nblocks = talloc.total_blocks(object)
        if object is None:
            nblocks -= self.initial_blocks
        self.assertEqual(nblocks, num_expected)

    def get_rodc_partial_attribute_set(self):
        """get a list of attributes for RODC replication"""
        partial_attribute_set = drsuapi.DsPartialAttributeSet()

        # we expect one block for the object
        self.check_blocks(partial_attribute_set, 1)

        attids = [1, 2, 3]
        partial_attribute_set.version = 1
        partial_attribute_set.attids     = attids
        partial_attribute_set.num_attids = len(attids)

        # we expect one block for the object, a structure, and a
        # reference to the array
        self.check_blocks(partial_attribute_set, 2)

        return partial_attribute_set

    def pas_test(self):
        pas = self.get_rodc_partial_attribute_set()
        self.check_blocks(pas, 2)
        req8 = drsuapi.DsGetNCChangesRequest8()
        self.check_blocks(req8, 1)

        # We expect the pas and req8, plus one block for each python object
        self.check_blocks(None, 5)
        req8.partial_attribute_set = pas
        if req8.partial_attribute_set.attids[1] != 2:
            raise Exception("Wrong value in attids[2]")
        # we now get an additional reference
        self.check_blocks(None, 6)

    def test_run(self):
        self.initial_blocks = talloc.total_blocks(None)
        self.check_blocks(None, 0)
        self.pas_test()
        self.check_blocks(None, 0)



# user management
#
# user unlock command
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba import ldb
from samba.auth import system_session
from samba.netcmd import Command, CommandError, Option
from samba.samdb import SamDB, SamDBError


class cmd_user_unlock(Command):
    """Unlock a user account.

    This command unlocks a user account in the Active Directory domain. The
    username specified on the command is the sAMAccountName. The username may
    also be specified using the --filter option.

    The command may be run from the root userid or another authorized userid.
    The -H or --URL= option can be used to execute the command against a remote
    server.

    Example:
    samba-tool user unlock user1 -H ldap://samba.samdom.example.com \\
        --username=Administrator --password=Passw0rd

    The example shows how to unlock a user account in the domain against a
    remote LDAP server. The -H parameter is used to specify the remote target
    server. The --username= and --password= options are used to pass the
    username and password of a user that exists on the remote server and is
    authorized to issue the command on that server.
"""

    synopsis = "%prog (<username>|--filter <filter>) [options]"

    takes_options = [
        Option("-H",
               "--URL",
               help="LDB URL for database or target server",
               type=str,
               metavar="URL",
               dest="H"),
        Option("--filter",
               help="LDAP Filter to set password on",
               type=str),
    ]

    takes_args = ["username?"]

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "versionopts": options.VersionOptions,
    }

    def run(self,
            username=None,
            sambaopts=None,
            credopts=None,
            versionopts=None,
            filter=None,
            H=None):
        if username is None and filter is None:
            raise CommandError("Either the username or '--filter' must be "
                               "specified!")

        if filter is None:
            filter = ("(&(objectClass=user)(sAMAccountName=%s))" % (
                ldb.binary_encode(username)))

        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp, fallback_machine=True)

        samdb = SamDB(url=H,
                      session_info=system_session(),
                      credentials=creds,
                      lp=lp)
        try:
            samdb.unlock_account(filter)
        except (SamDBError, ldb.LdbError) as msg:
            raise CommandError("Failed to unlock user '%s': %s" % (
                               username or filter, msg))



# domain management - domain functional_prep
#
# Copyright Matthias Dieter Wallnoefer 2009
# Copyright Andrew Kroeger 2009
# Copyright Jelmer Vernooij 2007-2012
# Copyright Giampaolo Lauria 2011
# Copyright Matthieu Patou <mat@matws.net> 2011
# Copyright Andrew Bartlett 2008-2015
# Copyright Stefan Metzmacher 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ldb
import samba.getopt as options
from samba.auth import system_session
from samba.dsdb import DS_DOMAIN_FUNCTION_2008, DS_DOMAIN_FUNCTION_2008_R2
from samba.netcmd import Command, CommandError, Option
from samba.netcmd.fsmo import get_fsmo_roleowner
from samba.samdb import SamDB

from samba import functional_level


class cmd_domain_functional_prep(Command):
    """Domain functional level preparation"""

    synopsis = "%prog [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
    }

    takes_options = [
        Option("-H", "--URL", help="LDB URL for database or target server", type=str,
               metavar="URL", dest="H"),
        Option("-q", "--quiet", help="Be quiet", action="store_true"),
        Option("-v", "--verbose", help="Be verbose", action="store_true"),
        Option("--function-level", type="choice", metavar="FUNCTION_LEVEL",
               choices=["2008_R2", "2012", "2012_R2", "2016"],
               help="The functional level to prepare for. Default is (Windows) 2016.",
               default="2016"),
        Option("--forest-prep", action="store_true",
               help="Run the forest prep (by default, both the domain and forest prep are run)."),
        Option("--domain-prep", action="store_true",
               help="Run the domain prep (by default, both the domain and forest prep are run).")
    ]

    def run(self, **kwargs):
        updates_allowed_overridden = False
        sambaopts = kwargs.get("sambaopts")
        credopts = kwargs.get("credopts")
        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)
        H = kwargs.get("H")
        function_level = kwargs.get("function_level")
        try:
            target_level = functional_level.string_to_level(function_level)
        except KeyError:
            raise CommandError(f"'{function_level}' is not known to Samba as an AD functional level")

        forest_prep = kwargs.get("forest_prep")
        domain_prep = kwargs.get("domain_prep")

        samdb = SamDB(url=H, session_info=system_session(), credentials=creds, lp=lp)

        # we're not going to get far if the config doesn't allow schema updates
        if lp.get("dsdb:schema update allowed") is None:
            lp.set("dsdb:schema update allowed", "yes")
            print("Temporarily overriding 'dsdb:schema update allowed' setting")
            updates_allowed_overridden = True

        if forest_prep is None and domain_prep is None:
            forest_prep = True
            domain_prep = True

        own_dn = ldb.Dn(samdb, samdb.get_dsServiceName())
        if forest_prep:
            master = get_fsmo_roleowner(samdb, str(samdb.get_schema_basedn()),
                                        'schema')
            if own_dn != master:
                raise CommandError("This server is not the schema master.")

        if domain_prep:
            domain_dn = samdb.domain_dn()
            infrastructure_dn = "CN=Infrastructure," + domain_dn
            master = get_fsmo_roleowner(samdb, infrastructure_dn,
                                        'infrastructure')
            if own_dn != master:
                raise CommandError("This server is not the infrastructure master.")

        exception_encountered = None

        if forest_prep and exception_encountered is None:
            samdb.transaction_start()
            try:
                from samba.forest_update import ForestUpdate
                forest = ForestUpdate(samdb, fix=True)

                forest.check_updates_iterator([11, 54, 79, 80, 81, 82, 83])
                forest.check_updates_functional_level(target_level,
                                                      DS_DOMAIN_FUNCTION_2008_R2,
                                                      update_revision=True)

                samdb.transaction_commit()
            except Exception as e:
                print("Exception: %s" % e)
                samdb.transaction_cancel()
                exception_encountered = e

        if domain_prep and exception_encountered is None:
            samdb.transaction_start()
            try:
                from samba.domain_update import DomainUpdate

                domain = DomainUpdate(samdb, fix=True)
                domain.check_updates_functional_level(target_level,
                                                      DS_DOMAIN_FUNCTION_2008,
                                                      update_revision=True)

                samdb.transaction_commit()
            except Exception as e:
                print("Exception: %s" % e)
                samdb.transaction_cancel()
                exception_encountered = e

        if updates_allowed_overridden:
            lp.set("dsdb:schema update allowed", "no")

        if exception_encountered is not None:
            raise CommandError('Failed to perform functional prep: %r' %
                               exception_encountered)



#!/usr/local/bin/python3.14
# Unix SMB/CIFS implementation.
# Copyright (C) Stefan Metzmacher 2020
# Copyright (C) 2020 Catalyst.Net Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

from samba.tests import DynamicTestCase

import ldb

from samba.tests.krb5.kdc_base_test import KDCBaseTest
from samba.tests.krb5.raw_testcase import KerberosCredentials
from samba.tests.krb5.rfc4120_constants import (
    AES256_CTS_HMAC_SHA1_96,
    ARCFOUR_HMAC_MD5,
    KDC_ERR_S_PRINCIPAL_UNKNOWN,
    NT_PRINCIPAL,
)

global_asn1_print = False
global_hexdump = False


@DynamicTestCase
class SpnTests(KDCBaseTest):
    test_account_types = {
        'computer': KDCBaseTest.AccountType.COMPUTER,
        'server': KDCBaseTest.AccountType.SERVER,
        'rodc': KDCBaseTest.AccountType.RODC
    }
    test_spns = {
        '2_part': 'ldap/{{account}}',
        '3_part_our_domain': 'ldap/{{account}}/{netbios_domain_name}',
        '3_part_our_realm': 'ldap/{{account}}/{dns_domain_name}',
        '3_part_not_our_realm': 'ldap/{{account}}/test',
        '3_part_instance': 'ldap/{{account}}:test/{dns_domain_name}'
    }

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls._mock_rodc_creds = None

    @classmethod
    def setUpDynamicTestCases(cls):
        for account_type_name, account_type in cls.test_account_types.items():
            for spn_name, spn in cls.test_spns.items():
                tname = f'{spn_name}_spn_{account_type_name}'
                targs = (account_type, spn)
                cls.generate_dynamic_test('test_spn', tname, *targs)

    def _test_spn_with_args(self, account_type, spn):
        target_creds = self._get_creds(account_type)
        spn = self._format_spn(spn, target_creds)

        sname = self.PrincipalName_create(name_type=NT_PRINCIPAL,
                                          names=spn.split('/'))

        client_creds = self.get_client_creds()
        tgt = self.get_tgt(client_creds)

        samdb = self.get_samdb()
        netbios_domain_name = samdb.domain_netbios_name()
        dns_domain_name = samdb.domain_dns_name()

        subkey = self.RandomKey(tgt.session_key.etype)

        etypes = (AES256_CTS_HMAC_SHA1_96, ARCFOUR_HMAC_MD5,)

        if account_type is self.AccountType.SERVER:
            ticket_etype = AES256_CTS_HMAC_SHA1_96
        else:
            ticket_etype = None
        decryption_key = self.TicketDecryptionKey_from_creds(
            target_creds, etype=ticket_etype)

        if (spn.count('/') > 1
                and (spn.endswith(netbios_domain_name)
                     or spn.endswith(dns_domain_name))
                and account_type is not self.AccountType.SERVER
                and account_type is not self.AccountType.RODC):
            expected_error_mode = KDC_ERR_S_PRINCIPAL_UNKNOWN
            check_error_fn = self.generic_check_kdc_error
            check_rep_fn = None
        else:
            expected_error_mode = 0
            check_error_fn = None
            check_rep_fn = self.generic_check_kdc_rep

        kdc_exchange_dict = self.tgs_exchange_dict(
            expected_crealm=tgt.crealm,
            expected_cname=tgt.cname,
            expected_srealm=tgt.srealm,
            expected_sname=sname,
            ticket_decryption_key=decryption_key,
            check_rep_fn=check_rep_fn,
            check_error_fn=check_error_fn,
            check_kdc_private_fn=self.generic_check_kdc_private,
            expected_error_mode=expected_error_mode,
            tgt=tgt,
            authenticator_subkey=subkey,
            kdc_options='0',
            expect_edata=False)

        self._generic_kdc_exchange(kdc_exchange_dict,
                                   cname=None,
                                   realm=tgt.srealm,
                                   sname=sname,
                                   etypes=etypes)

    def setUp(self):
        super().setUp()
        self.do_asn1_print = global_asn1_print
        self.do_hexdump = global_hexdump

    def _format_spns(self, spns, creds=None):
        return map(lambda spn: self._format_spn(spn, creds), spns)

    def _format_spn(self, spn, creds=None):
        samdb = self.get_samdb()

        spn = spn.format(netbios_domain_name=samdb.domain_netbios_name(),
                         dns_domain_name=samdb.domain_dns_name())

        if creds is not None:
            account_name = creds.get_username()
            spn = spn.format(account=account_name)

        return spn

    def _get_creds(self, account_type):
        spns = self._format_spns(self.test_spns.values())

        if account_type is self.AccountType.RODC:
            creds = self._mock_rodc_creds
            if creds is None:
                creds = self._get_mock_rodc_creds(spns)
                type(self)._mock_rodc_creds = creds
        else:
            creds = self.get_cached_creds(
                account_type=account_type,
                opts={
                    'spn': spns
                })

        return creds

    def _get_mock_rodc_creds(self, spns):
        rodc_ctx = self.get_mock_rodc_ctx()

        for spn in spns:
            spn = spn.format(account=rodc_ctx.myname)
            if spn not in rodc_ctx.SPNs:
                rodc_ctx.SPNs.append(spn)

        samdb = self.get_samdb()
        rodc_dn = ldb.Dn(samdb, rodc_ctx.acct_dn)

        msg = ldb.Message(rodc_dn)
        msg['servicePrincipalName'] = ldb.MessageElement(
            rodc_ctx.SPNs,
            ldb.FLAG_MOD_REPLACE,
            'servicePrincipalName')
        samdb.modify(msg)

        creds = KerberosCredentials()
        creds.guess(self.get_lp())
        creds.set_realm(rodc_ctx.realm.upper())
        creds.set_domain(rodc_ctx.domain_name)
        creds.set_password(rodc_ctx.acct_pass)
        creds.set_username(rodc_ctx.myname)
        creds.set_workstation(rodc_ctx.samname)
        creds.set_dn(rodc_dn)
        creds.set_spn(rodc_ctx.SPNs)

        res = samdb.search(base=rodc_dn,
                           scope=ldb.SCOPE_BASE,
                           attrs=['msDS-KeyVersionNumber'])
        kvno = int(res[0].get('msDS-KeyVersionNumber', idx=0))
        creds.set_kvno(kvno)

        keys = self.get_keys(creds)
        self.creds_set_keys(creds, keys)

        return creds


if __name__ == "__main__":
    global_asn1_print = False
    global_hexdump = False
    import unittest
    unittest.main()



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.dcerpc.registry."""

from samba.dcerpc import winreg
from samba.tests import RpcInterfaceTestCase


class WinregTests(RpcInterfaceTestCase):

    def setUp(self):
        super().setUp()
        self.conn = winreg.winreg("ncalrpc:", self.get_loadparm(),
                                  self.get_credentials())

    def get_hklm(self):
        return self.conn.OpenHKLM(None,
                                  winreg.KEY_QUERY_VALUE | winreg.KEY_ENUMERATE_SUB_KEYS)

    def test_hklm(self):
        handle = self.conn.OpenHKLM(None,
                                    winreg.KEY_QUERY_VALUE | winreg.KEY_ENUMERATE_SUB_KEYS)
        self.conn.CloseKey(handle)

    def test_getversion(self):
        handle = self.get_hklm()
        version = self.conn.GetVersion(handle)
        self.assertEqual(int, version.__class__)
        self.conn.CloseKey(handle)

    def test_getkeyinfo(self):
        handle = self.conn.OpenHKLM(None,
                                    winreg.KEY_QUERY_VALUE | winreg.KEY_ENUMERATE_SUB_KEYS)
        x = self.conn.QueryInfoKey(handle, winreg.String())
        self.assertEqual(9, len(x))  # should return a 9-tuple
        self.conn.CloseKey(handle)
